# FIXME: send fsgs as parameter to ValueConfigLoader instead
import json
import os

from fsbc.paths import Paths
from fsgs.amiga import whdload
from fsgs.amiga.amiga import Amiga
from fsgs.context import fsgs
from fsgs.filedatabase import FileDatabase
from fsgs.network import openretro_url_prefix
from fsgs.option import Option


class ValueConfigLoader(object):
    DB_VERSION_MAX = 1

    def __init__(self, uuid=""):
        self.config = {}
        self.options = {}
        self.viewport = []
        self.values = {}
        self.uuid = uuid
        self._file_list = None
        self._cue_sheets = None
        if uuid:
            self.options["database_url"] = "{0}/game/{1}".format(
                openretro_url_prefix(), uuid
            )

    def get_config(self):
        return self.config.copy()

    def load_values(self, values):
        if values.get("db_version"):
            try:
                version = int(values.get("db_version"))
            except ValueError:
                version = 10000
            if version > self.DB_VERSION_MAX:
                self.config["__config_name"] = (
                    "Unsupported Database "
                    "Version (Please upgrade FS-UAE Launcher)"
                )
                return self.get_config()

        self.values = values
        # from pprint import pprint
        # pprint(values)

        cd_based = False
        amiga_model = "A500"
        platform = self.values.get("platform", "Amiga")
        self.options["platform"] = platform.lower()

        platform = platform.lower()
        if platform == "amiga":
            pass
        elif platform == "cd32":
            amiga_model = "CD32"
            cd_based = True
            self.options["joystick_port_1_mode"] = "cd32 gamepad"
        elif platform == "cdtv":
            amiga_model = "CDTV"
            cd_based = True
        else:
            raise Exception("unknown platform")
        self.config["amiga_model"] = amiga_model

        # self.config["x_game_uuid"] = self.values["uuid"]

        self.viewport = []
        # game_node = self.root.find("game")
        # if game_node is not None:
        #     game_uuid = game_node.get("uuid", "")
        #     self.config["x_game_uuid"] = game_uuid
        #     self.load_game_info(game_uuid)

        # self.load_options())

        sort_list = []
        for key in values:
            if key == "chip_memory":
                # chip memory is checked at the end because of support for
                # the 1024+ value, and others keys can change amiga_model..
                sort_list.append(("x_chip_memory", key))
            else:
                sort_list.append((key, key))
        sort_list.sort()

        for dummy, key in sort_list:
            self.load_option(key, values[key])

        if self.viewport:
            self.options["viewport"] = ", ".join(self.viewport)

        if cd_based:
            if self.check_all_files():
                self.load_cdroms()
        else:
            self.load_floppies()

        self.load_hard_drives()

        # if not self.check_all_files():
        #     print(" -- some files are missing --")
        #     self.options["x_missing_files"] = "1"

        download_page = values.get("download_page", "")
        if download_page:
            self.options["download_page"] = download_page
        download_file = values.get("download_file", "")
        if download_file:
            self.options["download_file"] = download_file
        download_terms = values.get("download_terms", "")
        if download_terms:
            self.options["download_terms"] = download_terms
        download_notice = values.get("download_notice", "")
        if download_notice:
            self.options["download_notice"] = download_notice

        # for now, just copy all options to config without checking
        for key, value in self.options.items():
            self.config[key] = value

        if self.config.get(Option.X_WHDLOAD_ARGS):
            whdload.override_config(self.config)

        if (
            self.config.get("amiga_model", "") == "A500"
            and self.config.get("slow_memory") == "0"
        ):
            self.config["amiga_model"] = "A500/512K"
            self.config["slow_memory"] = ""

        self.set_name_and_uuid()
        self.contract_paths()
        return self.get_config()

    def check_all_files(self):
        file_list_json = self.values.get("file_list", "[]")
        file_list = json.loads(file_list_json)
        for file in file_list:
            if file["name"].endswith("/"):
                # this is a folder
                continue
            # print(repr(file))
            if not file["sha1"]:
                return False
            if not fsgs.file.find_by_sha1(file["sha1"]):
                return False
        return True

    def contract_paths(self):
        def fix(key):
            if self.config.get(key):
                self.config[key] = Paths.contract_path(
                    self.config.get(key), default_dir, force_real_case=False
                )

        default_dir = fsgs.amiga.get_floppies_dir()
        for i in range(Amiga.MAX_FLOPPY_DRIVES):
            fix("floppy_drive_{0}".format(i))
        for i in range(Amiga.MAX_FLOPPY_IMAGES):
            fix("floppy_image_{0}".format(i))

        default_dir = fsgs.amiga.get_cdroms_dir()
        for i in range(Amiga.MAX_CDROM_DRIVES):
            fix("cdrom_drive_{0}".format(i))
        for i in range(Amiga.MAX_CDROM_IMAGES):
            fix("cdrom_image_{0}".format(i))

    def set_name_and_uuid(self):
        # self.config["x_config_uuid"] = self.root.get("uuid", "")

        self.config["game_uuid"] = self.values.get("game_uuid", "")
        self.config["variant_uuid"] = self.values.get("variant_uuid", "")

        game_name = self.values.get("game_name", "")
        platform_name = self.values.get("platform", "")
        variant_name = self.values.get("variant_name", "")
        parts = []
        if platform_name:
            parts.append(platform_name)
        if variant_name:
            parts.append(variant_name)
        if game_name and variant_name:
            config_name = "{0} ({1})".format(game_name, ", ".join(parts))
            self.config["__config_name"] = config_name
        else:
            self.config["__config_name"] = self.values.get("config_name", "")

    def load_option(self, key, value):
        model = self.options.get("amiga_model", "")
        if key in ["variant_viewport", "viewport"]:
            if "=" in value:
                parts = value.split(",")
                for i in range(len(parts)):
                    parts[i] = parts[i].split("#", 1)[0].strip()
                    # parts[i] = parts[i].replace("=", "=>")
                    # parts[i] = parts[i].replace("==>", "=>")
                value = ", ".join(parts)
                while "  " in value:
                    value = value.replace("  ", " ")
            else:
                value = "* * * * = " + value
            self.viewport.append(value)
        # if key.startswith("viewport_"):
        #     parts = key.split("_")
        #     if len(parts) == 5:
        #         parts = parts[1:]
        #         value = " ".join(parts) + " => " + value
        #         value = value.replace("x", "*")
        #         self.viewport.append(value)
        elif key == "hd_startup":
            self.options["hd_startup"] = value
            if whdload.should_disable_drive_click():
                self.options[Option.FLOPPY_DRIVE_VOLUME_EMPTY] = "0"
        elif key == "whdload_args":
            self.options[Option.X_WHDLOAD_ARGS] = value
        elif key == "whdload_quit_key":
            self.options[Option.WHDLOAD_QUIT_KEY] = value
        elif key == "hdinst_args":
            self.options["x_hdinst_args"] = value
            if whdload.should_disable_drive_click():
                self.options[Option.FLOPPY_DRIVE_VOLUME_EMPTY] = "0"
        elif key == "hd_requirements":
            self.options["hd_requirements"] = value
        elif key == "save_disk":
            self.options["save_disk"] = value
        elif key == "whdload_version":
            self.options["x_whdload_version"] = value
        # elif key == "whdload_icon":
        #     self.options["__whdload_icon"] = value
        elif key == "kickstart":
            if value == "1.2":
                self.options["amiga_model"] = "A1000"
            elif value == "2.0":
                if model in ["A500+", "A600"]:
                    pass
                else:
                    self.options["amiga_model"] = "A600"
            elif value == "2.0+":
                if model in [
                    "A500+",
                    "A600",
                    "A1200",
                    "A1200/020",
                    "A3000",
                    "A4000/040",
                ]:
                    pass
                else:
                    self.options["amiga_model"] = "A600"
            elif value in ["3.0+", "3.1", "3.1+"]:
                if model in ["A1200", "A1200/020", "A3000", "A4000/040"]:
                    pass
                else:
                    self.options["amiga_model"] = "A1200"
            elif value == "AROS":
                self.options["kickstart_file"] = "internal"
            else:
                # FIXME: print warning
                pass
        elif key == "chipset":
            if value == "ECS":
                if model in ["A500+", "A600", "A3000"]:
                    pass
                else:
                    self.options["amiga_model"] = "A600"
            elif value == "AGA":
                if model in ["A1200", "A1200/020", "A4000/040"]:
                    pass
                else:
                    self.options["amiga_model"] = "A1200"
        elif key == "cpu":
            if value == "68020+" or value == "68020":
                if model in ["A1200", "A1200/020", "A3000", "A4000/040"]:
                    pass
                else:
                    self.options["amiga_model"] = "A1200"
        elif key == "fast_memory":
            ivalue = int(value)
            if ivalue > 8192:
                self.options["zorro_iii_memory"] = value
                self.options["amiga_model"] = "A1200/020"
            else:
                self.options["fast_memory"] = value
        elif key == "chip_memory":
            model = self.options.get("amiga_model", "")
            if value == "1024+":
                if model in ["A1200", "A1200/020", "A4000/040"]:
                    pass
                else:
                    self.options["chip_memory"] = "1024"
            elif value == "2048+":
                if model in ["A1200", "A1200/020", "A4000/040"]:
                    pass
                else:
                    # self.options["amiga_model"] = "A1200"
                    self.options["chip_memory"] = "2048"
            else:
                self.options["chip_memory"] = value
        elif key == "video_standard":
            if value == "NTSC":
                self.options["ntsc_mode"] = "1"
        elif key == "cracktro":
            # FIXME: handle
            pass
        elif key in [
            "joystick_port_0_mode",
            "joystick_port_1_mode",
            "joystick_port_2_mode",
            "joystick_port_3_mode",
            "joystick_port_4_mode",
        ]:
            self.load_joystick_port_x_mode_option(key, value)
        elif key in [
            "amiga_model",
            "accuracy",
            "cdrom_drive_0_delay",
            "floppy_drive_count",
            "slow_memory",
            "front_sha1",
            "screen1_sha1",
            "screen2_sha1",
            "screen3_sha1",
            "screen4_sha1",
            "screen5_sha1",
            "title_sha1",
            "year",
            "publisher",
            "developer",
            "hol_url",
            "lemon_url",
            "wikipedia_url",
            "mobygames_url",
            "whdload_url",
            "amigamemo_url",
            "longplay_url",
            "thelegacy_url",
            "homepage_url",
            "languages",
            "dongle_type",
        ]:
            self.options[key] = value
        elif key == "requirements":
            if "wb" in value.lower():
                self.options["hard_drive_0"] = "hd://template/workbench/DH0"
                self.options["hard_drive_0_priority"] = "6"
            elif "hd" in value.lower():
                self.options["hard_drive_0"] = "hd://template/empty/DH0"
        elif key == "players":
            self.options["players"] = value
        elif key == "protection":
            self.options["protection"] = value
        elif key == "game_notice":
            self.options["x_game_notice"] = value
        elif key == "variant_notice":
            self.options["x_variant_notice"] = value
        elif key == "variant_warning":
            self.options["x_variant_warning"] = value
        elif key == "variant_error":
            self.options["x_variant_error"] = value
        elif key == "joy_emu_conflict":
            self.options["x_joy_emu_conflict"] = value
            # elif key == "languages":
            #     self.options["x_languages"] = value

    def load_joystick_port_x_mode_option(self, key, value):
        value = value.lower()
        if "," not in value:
            self.options[key] = value
            return
        parts = value.split(",")
        assert "=" not in parts[0]
        self.options[key] = parts[0]
        port = ["0", "1", "2", "3", "4"].index(key[14])
        for part in parts[1:]:
            k, v = part.split("=")
            k = k.strip()
            v = v.strip()
            if not v.startswith("action_"):
                v = "action_" + v
            k = "joystick_port_{0}_{1}".format(port, k)
            self.options[k] = v

    def get_file_list(self):
        if self._file_list is None:
            file_list_json = self.values.get("file_list", "[]")
            self._file_list = json.loads(file_list_json)
        return self._file_list

    def get_cue_sheets(self):
        if self._cue_sheets is None:
            cue_sheets_json = self.values.get("cue_sheets", "[]")
            self._cue_sheets = json.loads(cue_sheets_json)
        return self._cue_sheets

    def build_media_list(self, floppies=False, cds=False, hds=False):
        media_list = []
        added = set()
        for file_item in self.get_file_list():
            name = file_item["name"]
            url = file_item.get("url", "")

            if name.startswith("DH0/"):
                if hds:
                    # p = os.path.join(self.path, "HardDrive")
                    p = "hd://game/" + self.uuid + "/DH0"
                    if p in added:
                        # already added
                        continue
                    added.add(p)
                    # FIXME: hack for now
                    sha1 = self.values.get("dh0_sha1", "")
                    media_list.append((p, sha1))
                else:
                    continue

            sha1 = file_item["sha1"]
            base, ext = os.path.splitext(name)
            ext = ext.lower()

            if hds:
                if ext not in [".zip"]:
                    continue
            elif cds:
                if ext not in [".cue", ".iso"]:
                    continue
                if "(Track" in base:
                    # probably part of a split multi-track cue
                    continue
            elif floppies:
                if ext not in [".adf", ".adz", ".dms", ".ipf"]:
                    continue

            path = ""
            found_sha1 = ""
            if sha1:
                print(sha1)
                file = FileDatabase.get_instance().find_file(sha1=sha1)
                if file:
                    found_sha1 = sha1
                    path = file["path"]
            if url and not path:
                path = url
                found_sha1 = sha1

            if path:
                media_list.append((path, found_sha1))
            else:
                pass
                # return False
                #  FIXME: handle it with a visible error message
                # raise Exception("could not find file " + repr(name))
        return media_list

    def load_floppies_from_floppy_list(self):
        media_list = []
        for item in self.values.get("floppy_list").split(","):
            name, sha1 = item.split(":")
            name = name.strip()
            sha1 = sha1.strip()
            path = "sha1://{0}/{1}/{2}".format(sha1, "", name)
            media_list.append((path, sha1))
        return media_list

    def load_floppies(self):
        floppy_drive_count = 4
        if "floppy_drive_count" in self.options:
            try:
                floppy_drive_count = int(self.options["floppy_drive_count"])
            except ValueError:
                floppy_drive_count = 1
            floppy_drive_count = max(0, min(4, floppy_drive_count))

        if self.values.get("floppy_list", ""):
            media_list = self.load_floppies_from_floppy_list()
        else:
            media_list = self.build_media_list(floppies=True)

        for i, values in enumerate(media_list):
            path, sha1 = values
            if i < floppy_drive_count:
                self.config["floppy_drive_{0}".format(i)] = path
                self.config["x_floppy_drive_{0}_sha1".format(i)] = sha1
            self.config["floppy_image_{0}".format(i)] = path
            self.config["x_floppy_image_{0}_sha1".format(i)] = sha1
        if floppy_drive_count < 4:
            self.config["floppy_drive_count"] = floppy_drive_count

    def cdrom_sha1_to_uri(self, sha1):
        for file_item in self.get_file_list():
            if file_item["sha1"] == sha1:
                return "game://{}/{}".format(self.uuid, file_item["name"])
        raise Exception("cdrom_sha1_to_uri: could not find " + sha1)

    def load_cdroms(self):
        # media_list = self.build_media_list(cds=True)
        cue_sheets = self.get_cue_sheets()
        if cue_sheets:
            return self.load_cdroms_from_cue_sheets(cue_sheets)
        cue = False
        ccd = False
        # iso = False
        for file_item in self.get_file_list():
            if file_item["name"].endswith(".cue"):
                cue = True
            if file_item["name"].endswith(".ccd"):
                ccd = True
                # if file_item["name"].endswith(".iso"):
                #     iso = True
        media_list = []
        if ccd:
            ext = ".ccd"
        elif cue:
            ext = ".cue"
        else:
            ext = ".iso"
        for file_item in self.get_file_list():
            if file_item["name"].endswith(ext):
                media_list.append((file_item["name"], file_item["sha1"]))
        print("load_cdroms media_list =", media_list)
        for i, values in enumerate(media_list):
            path, sha1 = values
            path = self.cdrom_sha1_to_uri(sha1)
            if i < 1:
                self.config["cdrom_drive_{0}".format(i)] = path
                self.config["x_cdrom_drive_{0}_sha1".format(i)] = sha1
            self.config["cdrom_image_{0}".format(i)] = path
            self.config["x_cdrom_image_{0}_sha1".format(i)] = sha1

    def load_cdroms_from_cue_sheets(self, cue_sheets):
        for i, item in enumerate(cue_sheets):
            path = "game://{}/{}".format(self.uuid, item["name"])
            if i < 1:
                self.config["cdrom_drive_{0}".format(i)] = path
            self.config["cdrom_image_{0}".format(i)] = path

    def load_hard_drives(self):
        media_list = self.build_media_list(hds=True)
        print(media_list)
        for i, values in enumerate(media_list):
            path, sha1 = values
            self.config["hard_drive_{0}".format(i)] = path
            self.config["x_hard_drive_{0}_sha1".format(i)] = sha1



from __future__ import absolute_import
from __future__ import print_function

from ..Block import Block

class BadBlockBlock(Block):
  def __init__(self, blkdev, blk_num=0):
    Block.__init__(self, blkdev, blk_num, chk_loc=2, is_type=Block.BADB)
  
  def create(self, block_pairs, host_id, size=128, next=0):
    Block.create(self)
    self.size = size
    self.host_id = host_id
    self.next = next
    self.block_pairs = block_pairs
      
  def write(self):
    self._create_data()
    
    self._put_long(1, self.size)
    self._put_long(3, self.host_id)
    self._put_long(4, self.next)
    
    # block_pairs: bad, good, bad, good ...
    off = 6
    for b in self.block_pairs:
      self._put_long(off, b)
      off += 1
    
    Block.write(self)
  
  def read(self):
    Block.read(self)
    if not self.valid:
      return False
    
    self.size = self._get_long(1)
    self.host_id = self._get_long(3)
    self.next = self._get_long(4)
    
    self.block_pairs = []
    off = 6
    while off < self.block_longs:
      b = self._get_long(off)
      if b == 0 or b == 0xffffffff:
        break
      self.block_pairs.append(b)
    
    return self.valid
  
  def dump(self):
    Block.dump(self, "RDBlock")
    
    print(" size:           %d" % self.size)
    print(" host_id:        %d" % self.host_id)
    print(" next:           %s" % self._dump_ptr(self.next))
    n = len(self.block_pairs) / 2
    o = 0
    for i in xrange(n):
      print(" bad=%d good=%d" % (self.block_pairs[o], self.block_pairs[o+1]))
      o += 2



import hashlib
import os

from fsgs import Option
from fsgs.drivers.mednafendriver import MednafenDriver
from fsgs.platform import PlatformHandler
from fsgs.platforms.loader import SimpleLoader


class SuperNintendoPlatformHandler(PlatformHandler):
    PLATFORM_NAME = "Super Nintendo"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return SuperNintendoLoader(fsgs)

    def get_runner(self, fsgs):
        return SuperNintendoMednafenDriver(fsgs)


class SuperNintendoLoader(SimpleLoader):
    pass


SNES_CONTROLLER = {
    "type": "gamepad",
    "description": "Gamepad",
    "mapping_name": "supernintendo",
}
SNES_PORTS = [
    {
        "description": "Port 1",
        "types": [SNES_CONTROLLER],
        "type_option": "snes_port_1_type",
        "device_option": "snes_port_1",
    },
    {
        "description": "Port 2",
        "types": [SNES_CONTROLLER],
        "type_option": "snes_port_2_type",
        "device_option": "snes_port_2",
    },
]


class SuperNintendoMednafenDriver(MednafenDriver):
    PORTS = SNES_PORTS

    def __init__(self, fsgc):
        super().__init__(fsgc)
        self.helper = SuperNintendoHelper(self.options)

    def prepare(self):
        print("[SNES] Mednafen SNES driver preparing...")
        super().prepare()
        self.set_mednafen_aspect(4, 3)
        # We do aspect calculation separately. Must not be done twice.
        self.emulator.args.extend(["-snes.correct_aspect", "0"])
        # FIXME: Input ports configuration
        # FIXME: SNES model
        # ROM path must be added at the end of the argument list
        self.emulator.args.append(self.helper.prepare_rom(self))

    # def mednafen_aspect_ratio(self):
    #     return 4.0 / 3.0

    def mednafen_input_mapping(self, port):
        n = port + 1
        return {
            "A": "snes.input.port{}.gamepad.a".format(n),
            "B": "snes.input.port{}.gamepad.b".format(n),
            "X": "snes.input.port{}.gamepad.x".format(n),
            "Y": "snes.input.port{}.gamepad.y".format(n),
            "L": "snes.input.port{}.gamepad.l".format(n),
            "R": "snes.input.port{}.gamepad.r".format(n),
            "UP": "snes.input.port{}.gamepad.up".format(n),
            "DOWN": "snes.input.port{}.gamepad.down".format(n),
            "LEFT": "snes.input.port{}.gamepad.left".format(n),
            "RIGHT": "snes.input.port{}.gamepad.right".format(n),
            "SELECT": "snes.input.port{}.gamepad.select".format(n),
            "START": "snes.input.port{}.gamepad.start".format(n),
        }

    def mednafen_system_prefix(self):
        return "snes"

    def game_video_par(self):
        # These may not be entirely correct...
        # if self.is_pal():
        #     return (4 / 3) / (256 / 239)
        # else:
        #     return (4 / 3) / (256 / 224)
        size = self.game_video_size()
        return (4 / 3) / (size[0] / size[1])

    def game_video_size(self):
        if self.is_pal():
            size = (256, 239)
        else:
            size = (256, 224)
        return size

    def get_game_file(self, config_key="cartridge_slot"):
        return None


class SuperNintendoHelper:
    def __init__(self, options):
        self.options = options

    # FIXME: Shared, move into common module (find all occurrences)
    def prepare_rom(self, driver):
        file_uri = self.options[Option.CARTRIDGE_SLOT]
        input_stream = driver.fsgc.file.open(file_uri)
        _, ext = os.path.splitext(file_uri)
        return self.prepare_rom_with_stream(driver, input_stream, ext)

    # FIXME: Shared, move into common module (find all occurrences)
    def prepare_rom_with_stream(self, driver, input_stream, ext):
        sha1_obj = hashlib.sha1()
        path = driver.temp_file("rom" + ext).path
        with open(path, "wb") as f:
            while True:
                data = input_stream.read(65536)
                if not data:
                    break
                f.write(data)
                sha1_obj.update(data)
        new_path = os.path.join(
            os.path.dirname(path), sha1_obj.hexdigest()[:8].upper() + ext
        )
        os.rename(path, new_path)
        return new_path



# def xx_on_resize(self):
#     if self.layout:
#         self.layout.set_size(self.get_size())
#         self.layout.update()


class Element:
    def __init__(self, parent, delay_create=False):
        self.parent = parent
        self.layout = None
        self.position = (0, 0)
        self.size = (0, 0)
        # if not delay_create:
        #     self.on_create()

    # @classmethod
    # def init(cls, obj):
    #     obj.on_resize = xx_on_resize

    def is_visible(self):
        return True

    def create(self):
        self.on_create()
        return self

    def get_position_base(self):
        base = self.get_position_base()
        pos = self.get_position()
        return base[0] + pos[0], base[1] + pos[1]

    def on_create(self):
        pass

    def get_real_parent(self):
        return self

    def get_size(self):
        return self.size

    def get_min_size(self):
        return 0, 0

    def get_position(self):
        return self.position

    def set_position(self, position):
        self.position = position

    def set_position_and_size(self, position, size):
        self.set_position(position)
        self.set_size(size)

    def set_size(self, size):
        self.size = size
        if self.layout:
            self.layout.set_size(size)
        self.on_resize()

    def on_resize(self):
        if self.layout:
            self.layout.set_size(self.get_size())
            self.layout.update()


class LightElement(Element):
    def __init__(self, parent):
        Element.__init__(self, parent)

    def get_real_parent(self):
        return self.parent.get_real_parent()



import hashlib
import json
import os
import shutil
import tempfile
import traceback
import zlib
from typing import Dict, List
from urllib.parse import unquote

from fsbc.paths import Paths
from fsbc.resources import Resources
from fsbc.task import TaskFailure, current_task
from fsgs.amiga import whdload
from fsgs.amiga.adffileextractor import ADFFileExtractor
from fsgs.amiga.amiga import Amiga
from fsgs.amiga.configwriter import ConfigWriter
from fsgs.amiga.fsuae import FSUAE
from fsgs.amiga.rommanager import ROMManager
from fsgs.amiga.roms import CD32_FMV_ROM, PICASSO_IV_74_ROM
from fsgs.amiga.workbench import WorkbenchExtractor
from fsgs.archive import Archive
from fsgs.download import Downloader
from fsgs.drivers.gamedriver import GameDriver
from fsgs.FSGSDirectories import FSGSDirectories
from fsgs.GameChangeHandler import GameChangeHandler
from fsgs.GameNameUtil import GameNameUtil
from fsgs.knownfiles import (
    ACTION_REPLAY_MK_II_2_14_MOD_ROM,
    ACTION_REPLAY_MK_II_2_14_ROM,
    ACTION_REPLAY_MK_III_3_17_MOD_ROM,
    ACTION_REPLAY_MK_III_3_17_ROM,
)
from fsgs.network import is_http_url
from fsgs.option import Option
from fsgs.res import gettext


class LaunchHandler(object):
    def __init__(self, fsgs, config_name, config, game_paths, temp_dir=""):
        self.fsgs = fsgs
        self.fsgc = fsgs
        self.config_name = config_name
        self.config = config.copy()

        for remove_key in [
            "database_username",
            "database_password",
            "database_username",
            "database_email",
            "database_auth",
            "device_id",
        ]:
            if remove_key in self.config:
                del self.config[remove_key]

        # make sure FS-UAE does not load other config files (Host.fs-uae)
        self.config["end_config"] = "1"

        self.game_paths = game_paths
        self.hd_requirements = set()
        for req in (
            self.config.get("hd_requirements", "").replace(",", ";").split(";")
        ):
            req = req.strip()
            if req:
                self.hd_requirements.add(req)
        self.setpatch_installed = False
        # self.stop_flag = False

        self.temp_dir = temp_dir
        self.change_handler = None

        self.use_relative_paths = (
            self.config.get(Option.RELATIVE_TEMP_FEATURE, "") == "1"
        )

    @property
    def stop_flag(self):
        return current_task.stop_flag

    def on_progress(self, progress):
        # method can be overridden / replaced in instances
        pass

    def on_complete(self):
        # method can be overridden / replaced in instances
        pass

    def prepare(self):
        print("LaunchHandler.prepare")
        if not self.temp_dir:
            run_dir = os.environ.get("FS_GAMESYS_RUN_DIR", "")
            if run_dir:
                if not os.path.isdir(run_dir):
                    os.makedirs(run_dir)
                if len(os.listdir(run_dir)) != 0:
                    raise Exception("FS_GAMESYS_RUN_DIR is not empty")
                self.temp_dir = run_dir
            else:
                self.temp_dir = tempfile.mkdtemp(prefix="fs-uae-")
        print("temp dir", self.temp_dir)
        self.config["floppies_dir"] = self.temp_dir
        print("state dir", self.get_state_dir())
        self.config["state_dir"] = self.get_state_dir()
        self.config["save_states_dir"] = ""
        self.config["floppy_overlays_dir"] = ""
        self.config["flash_memory_dir"] = ""
        self.change_handler = GameChangeHandler(self.temp_dir)

        self.config["cdroms_dir"] = FSGSDirectories.get_cdroms_dir()
        self.config["configurations_dir"] = (
            FSGSDirectories.get_configurations_dir()
        )
        self.config["controllers_dir"] = FSGSDirectories.get_controllers_dir()
        self.config["hard_drives_dir"] = FSGSDirectories.get_hard_drives_dir()
        self.config["kickstarts_dir"] = FSGSDirectories.get_kickstarts_dir()
        self.config["save_states_dir"] = FSGSDirectories.get_save_states_dir()
        self.config["themes_dir"] = FSGSDirectories.get_themes_dir()

        self.prepare_roms()
        if self.stop_flag:
            return
        self.prepare_floppies()
        if self.stop_flag:
            return
        self.prepare_cdroms()
        if self.stop_flag:
            return
        self.prepare_hard_drives()
        if self.stop_flag:
            return
        self.copy_hd_files()
        if self.stop_flag:
            return
        self.init_changes()
        if self.stop_flag:
            return
        self.prepare_theme()
        if self.stop_flag:
            return
        self.prepare_extra_settings()

    def run_sequence(self, start=True, cleanup=True):
        print("LaunchHandler.run_sequence")
        self.prepare()

        if not self.stop_flag:
            # too late to stop now...
            if start:
                self.run()
            self.update_changes()

        if cleanup:
            self.cleanup()

        print("calling LaunchHandler.on_complete")
        self.on_complete()

    def prepare_roms(self):
        print("LaunchHandler.prepare_roms")
        current_task.set_progress(gettext("Preparing kickstart ROMs..."))
        amiga_model = self.config.get("amiga_model", "A500")
        model_config = Amiga.get_model_config(amiga_model)

        roms = [("kickstart_file", model_config["kickstarts"])]
        if self.config["kickstart_ext_file"] or model_config["ext_roms"]:
            # not all Amigas have extended ROMs
            roms.append(("kickstart_ext_file", model_config["ext_roms"]))
        if amiga_model.lower() == "cd32/fmv":
            roms.append(("fvm_rom", [CD32_FMV_ROM]))

        if self.config["graphics_card"].lower().startswith("picasso-iv"):
            roms.append(("graphics_card_rom", [PICASSO_IV_74_ROM]))

        if self.config["accelerator"].lower() == "cyberstorm-ppc":
            roms.append(("accelerator_rom", ["cyberstormppc.rom"]))

        if self.config["freezer_cartridge"] == "action-replay-2":
            # Ideally, we would want to recognize ROMs based on zeroing the
            # first four bytes, but right now we simply recognize a common
            # additional version. freezer_cartridge_rom isn't a real option,
            # we just want to copy the rom file and let FS-UAE find it
            roms.append(
                (
                    "[freezer_cartridge]",
                    [
                        ACTION_REPLAY_MK_II_2_14_ROM.sha1,
                        ACTION_REPLAY_MK_II_2_14_MOD_ROM.sha1,
                    ],
                )
            )
        elif self.config["freezer_cartridge"] == "action-replay-3":
            roms.append(
                (
                    "[freezer_cartridge]",
                    [
                        ACTION_REPLAY_MK_III_3_17_ROM.sha1,
                        ACTION_REPLAY_MK_III_3_17_MOD_ROM.sha1,
                    ],
                )
            )

        use_temp_kickstarts_dir = False

        for config_key, default_roms in roms:
            print("[ROM]", config_key, default_roms)
            src = self.config[config_key]
            print(src)
            if not src:
                for sha1 in default_roms:
                    print("[ROM] Trying", sha1)
                    if is_sha1(sha1):
                        rom_src = self.fsgs.file.find_by_sha1(sha1)
                        if rom_src:
                            src = rom_src
                            print("[ROM] Found", rom_src)
                            break
                    else:
                        # roms_dir = FSGSDirectories.get_kickstarts_dir()
                        # src = os.path.join(roms_dir, sha1)
                        # if os.path.exists(src):
                        #     break
                        # loop up file in roms dir instead
                        src = sha1
            elif src == "internal":
                continue
            elif src:
                src = Paths.expand_path(src)
            if not src:
                raise TaskFailure(
                    gettext(
                        "Did not find required Kickstart or "
                        "ROM for {}. Wanted one of these files: {}".format(
                            config_key, repr(default_roms)
                        )
                    )
                )

            dest = os.path.join(self.temp_dir, os.path.basename(src))

            def lookup_rom_from_src(src):
                parts = src.split(":", 1)
                if len(parts) == 2 and len(parts[0]) > 1:
                    # src has a scheme (not a Windows drive letter). Assume
                    # we can find this file.
                    return src
                archive = Archive(src)
                if archive.exists(src):
                    return src
                dirs = [self.fsgs.amiga.get_kickstarts_dir()]
                for dir_ in dirs:
                    path = os.path.join(dir_, src)
                    print("[ROM] Checking", repr(path))
                    archive = Archive(path)
                    if archive.exists(path):
                        return path
                return None

            org_src = src
            src = lookup_rom_from_src(src)
            if not src and org_src == "cyberstormppc.rom":
                src = lookup_rom_from_src(
                    "ralphschmidt-cyberstorm-ppc-4471.rom"
                )
                if not src:
                    for (
                        dir_
                    ) in FSGSDirectories.get_amiga_forever_directories():
                        path = os.path.join(
                            dir_,
                            "Shared",
                            "rom",
                            "ralphschmidt-cyberstorm-ppc-4471.rom",
                        )
                        if os.path.exists(path):
                            src = path
                            print("[ROM] Found", path)
                            break
                        else:
                            print("[ROM] Trying", path)
            stream = None
            # FIXME: prepare_roms should be rewritten, it's kind of crap.
            # Rom patching and decryption should be handled differently. Should
            # use file database filters, and decryption via rom.key should only
            # be supported when using uncompressed files directly on disk.
            if not src or not os.path.exists(src):
                try:
                    stream = self.fsgs.file.open(src)
                    if stream is None:
                        raise FileNotFoundError(src)
                except FileNotFoundError:
                    raise TaskFailure(
                        gettext(
                            "Cannot find required ROM file: {name}".format(
                                name=repr(org_src)
                            )
                        )
                    )
            with open(dest, "wb") as f:
                if stream:
                    print("[ROM] From stream => {}".format(dest))
                    rom = {}
                    rom["data"] = stream.read()
                    rom["sha1"] = hashlib.sha1(rom["data"]).hexdigest()
                    ROMManager.patch_rom(rom)
                    f.write(rom["data"])
                else:
                    archive = Archive(src)
                    ROMManager.decrypt_archive_rom(archive, src, file=f)
                if use_temp_kickstarts_dir:
                    self.config[config_key] = os.path.basename(src)
                else:
                    self.config[config_key] = dest
        if use_temp_kickstarts_dir:
            self.config["kickstarts_dir"] = self.temp_dir

    @staticmethod
    def expand_default_path(src, default_dir):
        if "://" in src:
            return src, None
        src = Paths.expand_path(src, default_dir)
        archive = Archive(src)
        # if not archive.exists(src):
        #     dirs = [default_dir]
        #     for dir in dirs:
        #         path = os.path.join(dir, src)
        #         print("checking", repr(path))
        #         archive = Archive(path)
        #         if archive.exists(path):
        #         #if os.path.exists(path):
        #             src = path
        #             break
        #     else:
        #         raise Exception("Cannot find path for " + repr(src))
        return src, archive

    def prepare_floppy(self, key):
        src = self.config.get(key, "").strip()
        if not src:
            return

        src, archive = self.expand_default_path(
            src, self.fsgs.amiga.get_floppies_dir()
        )
        dst_name = os.path.basename(src)
        current_task.set_progress(dst_name)

        if self.config["writable_floppy_images"] == "1" and os.path.isfile(
            src
        ):
            # the config value directly refers to a local file, and the config
            # value already refers to the file, but since we may have
            # changed floppy_dir and the path may be relative, we set the
            # resolved path directly
            self.config[key] = src
        else:
            dst = os.path.join(self.temp_dir, dst_name)
            self.fsgs.file.copy_game_file(src, dst)
            self.config[key] = os.path.basename(dst)

    def prepare_floppies(self):
        print("LaunchHandler.copy_floppies")
        current_task.set_progress(gettext("Preparing floppy images..."))
        # self.on_progress(gettext("Preparing floppy images..."))

        floppies = []
        for i in range(Amiga.MAX_FLOPPY_DRIVES):
            key = "floppy_drive_{0}".format(i)
            if self.config.get(key, ""):
                floppies.append(self.config[key])
            self.prepare_floppy(key)

        for i in range(Amiga.MAX_FLOPPY_IMAGES):
            key = "floppy_image_{0}".format(i)
            if self.config.get(key, ""):
                break
        else:
            print("floppy image list is empty")
            for j, floppy in enumerate(floppies):
                self.config["floppy_image_{0}".format(j)] = floppy

        max_image = -1
        for i in range(Amiga.MAX_FLOPPY_IMAGES):
            key = "floppy_image_{0}".format(i)
            self.prepare_floppy(key)
            if self.config.get(key, ""):
                max_image = i

        save_image = max_image + 1

        if self.config.get("save_disk", "") != "0":
            s = Resources("fsgs", "res").stream("amiga/adf_save_disk.dat")
            data = s.read()
            data = zlib.decompress(data)
            save_disk = os.path.join(self.temp_dir, "Save Disk.adf")
            with open(save_disk, "wb") as f:
                f.write(data)
            key = "floppy_image_{0}".format(save_image)
            self.config[key] = "Save Disk.adf"

    def prepare_cdroms(self):
        print("LaunchHandler.prepare_cdroms")
        if not self.config.get("cdrom_drive_count", ""):
            if self.config.get("cdrom_drive_0", "") or self.config.get(
                "cdrom_image_0", ""
            ):
                self.config["cdrom_drive_count"] = "1"

        cdrom_drive_0 = self.config.get("cdrom_drive_0", "")
        if cdrom_drive_0.startswith("game:"):
            scheme, dummy, game_uuid, name = cdrom_drive_0.split("/")
            file_list = self.get_file_list_for_game_uuid(game_uuid)
            for file_item in file_list:
                src = self.fsgs.file.find_by_sha1(file_item["sha1"])

                src, archive = self.expand_default_path(
                    src, self.fsgs.amiga.get_cdroms_dir()
                )
                dst_name = file_item["name"]
                current_task.set_progress(dst_name)

                dst = os.path.join(self.temp_dir, dst_name)
                self.fsgs.file.copy_game_file(src, dst)

            cue_sheets = self.get_cue_sheets_for_game_uuid(game_uuid)
            for cue_sheet in cue_sheets:
                # FIXME: Try to get this to work with the PyCharm type checker
                # noinspection PyTypeChecker
                with open(
                    os.path.join(self.temp_dir, cue_sheet["name"]), "wb"
                ) as f:
                    # noinspection PyTypeChecker
                    f.write(cue_sheet["data"].encode("UTF-8"))

            for i in range(Amiga.MAX_CDROM_DRIVES):
                key = "cdrom_drive_{0}".format(i)
                value = self.config.get(key, "")
                if value:
                    self.config[key] = os.path.join(
                        self.temp_dir, os.path.basename(value)
                    )

            for i in range(Amiga.MAX_CDROM_IMAGES):
                key = "cdrom_image_{0}".format(i)
                value = self.config.get(key, "")
                if value:
                    self.config[key] = os.path.join(
                        self.temp_dir, os.path.basename(value)
                    )

        cdroms = []
        for i in range(Amiga.MAX_CDROM_DRIVES):
            key = "cdrom_drive_{0}".format(i)
            if self.config.get(key, ""):
                cdroms.append(self.config[key])

        for i in range(Amiga.MAX_CDROM_IMAGES):
            key = "cdrom_image_{0}".format(i)
            if self.config.get(key, ""):
                break
        else:
            print("CD-ROM image list is empty")
            for j, cdrom in enumerate(cdroms):
                self.config["cdrom_image_{0}".format(j)] = cdrom

    def prepare_hard_drives(self):
        print("LaunchHandler.prepare_hard_drives")
        current_task.set_progress(gettext("Preparing hard drives..."))
        # self.on_progress(gettext("Preparing hard drives..."))
        for i in range(0, 10):
            self.prepare_hard_drive(i)

    def prepare_hard_drive(self, index):
        key = "hard_drive_{}".format(index)
        src = self.config.get(key, "")
        dummy, ext = os.path.splitext(src)
        ext = ext.lower()

        if is_http_url(src):
            name = src.rsplit("/", 1)[-1]
            name = unquote(name)
            self.on_progress(gettext("Downloading {0}...".format(name)))
            dest = os.path.join(self.temp_dir, name)
            Downloader.install_file_from_url(src, dest)
            src = dest
        elif src.startswith("hd://game/"):
            self.unpack_game_hard_drive(index, src)
            self.disable_save_states()
            return
        elif src.startswith("file_list:"):
            self.unpack_game_hard_drive(index, src)
            self.disable_save_states()
            return
        elif src.startswith("hd://template/workbench/"):
            self.prepare_workbench_hard_drive(index, src)
            self.disable_save_states()
            return
        elif src.startswith("hd://template/empty/"):
            self.prepare_empty_hard_drive(index, src)
            self.disable_save_states()
            return

        if ext in Archive.extensions:
            print("zipped hard drive", src)
            self.unpack_hard_drive(index, src)
            self.disable_save_states()

        elif src.endswith("HardDrive"):
            print("XML-described hard drive", src)
            self.unpack_hard_drive(index, src)
            self.disable_save_states()
        else:
            src = Paths.expand_path(src)
            self.config[key] = src

    def disable_save_states(self):
        # Save states cannot currently be used with temporarily created
        # hard drives, as HD paths are embedded into the save states, and
        # restoring the save state causes problems.

        if self.config.get("unsafe_save_states") == "1":
            return
        self.config["save_states"] = "0"

    def prepare_workbench_hard_drive(self, i, src):
        # dir_name = "DH{0}".format(i)
        dir_name = src.rsplit("/", 1)[-1]
        dir_path = os.path.join(self.temp_dir, dir_name)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        amiga_model = self.config.get("amiga_model", "A500")
        if (
            amiga_model.startswith("A1200")
            or amiga_model.startswith("A4000")
            or amiga_model.startswith("A3000")
        ):
            workbench = "Minimal Workbench v3.1"
        elif amiga_model == "A600":
            workbench = "Minimal Workbench v2.05"
        elif amiga_model == "A500+":
            workbench = "Minimal Workbench v2.04"
        else:
            workbench = "Minimal Workbench v1.3"

        print("Try to find pre-configured hard drive", workbench)
        src_dir = os.path.join(
            self.fsgs.amiga.get_hard_drives_dir(), workbench
        )
        if src_dir and os.path.exists(src_dir):
            print("found", src_dir)
            self.copy_folder_tree(src_dir, dir_path)
        else:
            print(" - not found -")
            raise Exception(
                "Did not found pre-configured hard drive " + repr(workbench)
            )

        self.config["hard_drive_{0}".format(i)] = dir_path

    def prepare_empty_hard_drive(self, i, src):
        dir_name = src.rsplit("/", 1)[-1]
        # dir_name = "DH{0}".format(i)
        dir_path = os.path.join(self.temp_dir, dir_name)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        self.config["hard_drive_{0}".format(i)] = dir_path

    def get_file_list_for_game_uuid(self, game_uuid):
        # FIXME: This is an ugly hack, we should already be told what
        # database to use.
        try:
            game_database = self.fsgs.get_game_database()
            values = game_database.get_game_values_for_uuid(game_uuid)
        except LookupError:
            try:
                game_database = self.fsgs.game_database("CD32")
                values = game_database.get_game_values_for_uuid(game_uuid)
            except LookupError:
                game_database = self.fsgs.game_database("CDTV")
                values = game_database.get_game_values_for_uuid(game_uuid)
        file_list = json.loads(values["file_list"])
        return file_list

    def get_cue_sheets_for_game_uuid(self, game_uuid) -> List[Dict]:
        # FIXME: This is an ugly hack, we should already be told what
        # database to use.
        try:
            game_database = self.fsgs.get_game_database()
            values = game_database.get_game_values_for_uuid(game_uuid)
        except LookupError:
            try:
                game_database = self.fsgs.game_database("CD32")
                values = game_database.get_game_values_for_uuid(game_uuid)
            except LookupError:
                game_database = self.fsgs.game_database("CDTV")
                values = game_database.get_game_values_for_uuid(game_uuid)
        if not values.get("cue_sheets", ""):
            return []
        return json.loads(values["cue_sheets"])

    def unpack_game_hard_drive(self, drive_index, src):
        print("unpack_game_hard_drive", drive_index, src)
        if src.startswith("file_list:"):
            _scheme, dummy, drive = src.split("/")
            file_list = json.loads(self.fsgs.config.get("file_list"))
        else:
            _scheme, dummy, dummy, game_uuid, drive = src.split("/")
            file_list = self.get_file_list_for_game_uuid(game_uuid)

        drive_prefix = drive + "/"
        dir_name = "DH{0}".format(drive_index)
        dir_path = os.path.join(self.temp_dir, dir_name)

        for file_entry in file_list:
            if self.stop_flag:
                return

            name = file_entry["name"]
            if not name.startswith(drive_prefix):
                continue

            # extract Amiga relative path and convert each path component
            # to host file name (where needed).

            # amiga_rel_path = name[len(drive_prefix) :]
            # print("amiga_rel_path", amiga_rel_path)
            # amiga_rel_parts = amiga_rel_path.split("/")
            # for i, part in enumerate(amiga_rel_parts):
            #     # part can be blank if amiga_rel_parts is a directory
            #     # (ending with /)
            #     if part:
            #         amiga_rel_parts[i] = amiga_filename_to_host_filename(part)
            # amiga_rel_path = "/".join(amiga_rel_parts)
            amiga_rel_path = amiga_path_to_host_path(name[len(drive_prefix) :])

            dst_file = os.path.join(dir_path, amiga_rel_path)
            print(repr(dst_file))
            if name.endswith("/"):
                os.makedirs(Paths.str(dst_file))
                continue
            if not os.path.exists(os.path.dirname(dst_file)):
                os.makedirs(os.path.dirname(dst_file))
            sha1 = file_entry["sha1"]

            # current_task.set_progress(os.path.basename(dst_file))
            current_task.set_progress(amiga_rel_path)
            self.fsgs.file.copy_game_file("sha1://{0}".format(sha1), dst_file)

            # src_file = self.fsgs.file.find_by_sha1(sha1)
            # if not os.path.exists(os.path.dirname(dst_file)):
            #     os.makedirs(os.path.dirname(dst_file))
            # stream = self.fsgs.file.open(src_file)
            # # archive = Archive(src_file)
            # # f = archive.open(src_file)
            # data = stream.read()
            # assert hashlib.sha1(data).hexdigest() == sha1
            # with open(dst_file, "wb") as out_file:
            #     out_file.write(data)

            # noinspection SpellCheckingInspection
            metadata = [
                "----rwed",
                " ",
                "2000-01-01 00:00:00.00",
                " ",
                "",
                "\n",
            ]
            if "comment" in file_entry:
                metadata[4] = self.encode_file_comment(file_entry["comment"])
            with open(dst_file + ".uaem", "wb") as out_file:
                out_file.write("".join(metadata).encode("UTF-8"))

        if self.use_relative_paths:
            self.config["hard_drive_{0}".format(drive_index)] = dir_name
        else:
            self.config["hard_drive_{0}".format(drive_index)] = dir_path

    @staticmethod
    def encode_file_comment(comment):
        result = []
        # raw = 0
        for c in comment:
            # if c == '%':
            #     result.append("%")
            #     raw = 2
            # elif raw:
            #     result.append(c)
            #     raw = raw - 1
            # else:
            #     result.append("%{0:x}".format(ord(c)))
            result.append("%{0:x}".format(ord(c)))
        return "".join(result)

    def unpack_hard_drive(self, i, src):
        if not src.startswith("$") and os.path.exists(src):
            # In case a relative path points to an existing file.
            pass
        else:
            src, _archive = self.expand_default_path(
                src, self.fsgs.amiga.get_hard_drives_dir()
            )

        dir_name = "DH{0}".format(i)
        dir_path = os.path.join(self.temp_dir, dir_name)

        self.unpack_archive(src, dir_path)
        self.config["hard_drive_{0}".format(i)] = dir_path

    # def create_devs_dir(self):
    #     devs_dir = os.path.join(dest_dir, "Devs")
    #     if not os.path.exists(devs_dir):
    #         os.makedirs(devs_dir)
    #
    # def create_fonts_dir(self):
    #     fonts_dir = os.path.join(dest_dir, "Fonts")
    #     if not os.path.exists(fonts_dir):
    #         os.makedirs(fonts_dir)

    def copy_hd_files(self):
        whdload_args = self.config.get("x_whdload_args", "").strip()
        hdinst_args = self.config.get("x_hdinst_args", "").strip()
        hd_startup = self.config.get("hd_startup", "").strip()

        if not whdload_args:
            # The WHDLoad override setting and config key does not quite
            # follow the usual semantics of configs/settings unfortunately, so
            # we really want whdload_quit_key to be cleared when not using
            # WHDLoad. Otherwise the emulator will try to quit everything with
            # the WHDLoad quit key (when overriden).
            self.config["whdload_quit_key"] = ""

        if not whdload_args and not hdinst_args and not hd_startup:
            return

        dest_dir = os.path.join(self.temp_dir, "DH0")
        if not self.config.get("hard_drive_0", ""):
            self.config["hard_drive_0"] = dest_dir
            self.config["hard_drive_0_label"] = "Workbench"

        print("copy_hd_files, dest_dir = ", dest_dir)

        s_dir = os.path.join(dest_dir, "S")
        if not os.path.exists(s_dir):
            os.makedirs(s_dir)
        libs_dir = os.path.join(dest_dir, "Libs")
        if not os.path.exists(libs_dir):
            os.makedirs(libs_dir)

        devs_dir = os.path.join(dest_dir, "Devs")
        if not os.path.exists(devs_dir):
            os.makedirs(devs_dir)
        fonts_dir = os.path.join(dest_dir, "Fonts")
        if not os.path.exists(fonts_dir):
            os.makedirs(fonts_dir)

        if hd_startup:
            self.config["hard_drive_0_priority"] = "6"
            # don't copy setpatch by default, at least not yet
            pass
        else:
            self.hd_requirements.add("setpatch")
            self.copy_setpatch(dest_dir)

        amiga_model = self.config.get("amiga_model", "A500").upper()
        if amiga_model in ["A500+", "A600"]:
            workbench_version = "2.04"
        elif amiga_model.startswith("A1200"):
            workbench_version = "3.0"
        elif amiga_model.startswith("A4000"):
            workbench_version = "3.0"
        else:
            workbench_version = None

        if "workbench" in self.hd_requirements:
            if not workbench_version:
                raise Exception(
                    "Unsupported workbench version for hd_requirements"
                )
            extractor = WorkbenchExtractor(self.fsgs)
            extractor.install_version(workbench_version, dest_dir)
            # install_workbench_files(self.fsgs, dest_dir, workbench_version)

        for req in self.hd_requirements:
            if "/" in req:
                # assume a specific workbench file
                extractor = WorkbenchExtractor(self.fsgs)
                extractor.install_version(
                    workbench_version,
                    dest_dir,
                    [req],
                    install_startup_sequence=False,
                )

        if whdload_args:
            self.copy_whdload_files(dest_dir, s_dir)
        elif hdinst_args:
            self.write_startup_sequence(s_dir, hdinst_args)
        elif hd_startup:
            self.write_startup_sequence(s_dir, hd_startup)

        if "xpkmaster.library" in self.hd_requirements:
            self.copy_xpkmaster_files(dest_dir)

        system_configuration_file = os.path.join(
            devs_dir, "system-configuration"
        )
        if not os.path.exists(system_configuration_file):
            with open(system_configuration_file, "wb") as f:
                f.write(system_configuration)

    def copy_xpkmaster_files(self, dest_dir):
        file_map = {
            "Libs/xpkmaster.library": "5bd19f9503b59c5d19bfe1c6a6e3b6e7c0e9eae2",
            "Libs/compressors/xpkCBR0.library": "a2a76c10cb06315e51990911fa050669cc89830d",
            "Libs/compressors/xpkDLTA.library": "ca64f89919c2869cb6fd75346b9a21245a6d04a8",
            "Libs/compressors/xpkDUKE.library": "8102d77ae0a3d64496436ee56a9c577c84b11992",
            "Libs/compressors/xpkFAST.library": "4599430f3baa302635a85b26a7b70a4116dc5f09",
            "Libs/compressors/xpkFRLE.library": "d0746429187ab38e886a47820203315f734e8d89",
            "Libs/compressors/xpkHFMN.library": "f4a8dbe69e386d87d9ab8cbaf7fc3881b358fdb2",
            "Libs/compressors/xpkHUFF.library": "f22d099b81d2d039c4fbb3fea47cc9700b01fecf",
            "Libs/compressors/xpkIMPL.library": "0e00bc18aec757d86f9831131c2236a704c175db",
            "Libs/compressors/xpkMASH.library": "1093133a17cb635c21c5532ae26ded83b6d359ce",
            "Libs/compressors/xpkNONE.library": "46dca6e2ff2176f7a12c52ea992a7a49ae5ef269",
            "Libs/compressors/xpkNUKE.library": "168bee97805be1f85b65f615e6931c4942caf4e4",
            "Libs/compressors/xpkRAKE.library": "50cd2a19bee1ebd6b54b31c5b9461d5a81a2e910",
            "Libs/compressors/xpkRLEN.library": "c81cf6d99de56faa6154cc81a9876517c2a8efc0",
            "Libs/compressors/xpkSHRI.library": "fc96d367e7b3409074841c7f145ab6daef0a6a4d",
            "Libs/compressors/xpkSMPL.library": "d140b4e87a2ff76cf94616d97fe7b127d128d973",
            "Libs/compressors/xpkSQSH.library": "2191b616abd4b2fb34c89ae243c35feea2b71104",
        }
        for rel_path, sha1 in file_map.items():
            # file_dest_dir = os.path.join(dest_dir, os.path.dirname(name))
            # self.install_whdload_file(sha1, file_dest_dir, value)
            self.install_whdload_file(sha1, dest_dir, rel_path)

    def copy_whdload_files(self, dest_dir, s_dir):
        whdload_args = self.config.get("x_whdload_args", "").strip()
        if not whdload_args:
            return
        print("[WHDLOAD] LaunchHandler.copy_whdload_files")
        if self.config.get(Option.WHDLOAD_PRELOAD, "") != "0":
            if " PRELOAD" not in whdload_args.upper():
                print("[WHDLOAD] Adding PRELOAD argument")
                whdload_args += " PRELOAD"

        current_task.set_progress(gettext("Preparing WHDLoad..."))
        # self.on_progress(gettext("Preparing WHDLoad..."))
        print("[WHDLOAD] copy_whdload_files, dest_dir = ", dest_dir)

        whdload_dir = ""
        slave_original_name = whdload_args.split(" ", 1)[0]
        slave = slave_original_name.lower()
        found_slave = False
        for dir_path, dir_names, file_names in os.walk(dest_dir):
            for name in file_names:
                if name.lower() == slave:
                    print("[WHDLOAD] Found", name)
                    found_slave = True
                    whdload_dir = dir_path[len(dest_dir) :]
                    whdload_dir = whdload_dir.replace("\\", "/")
                    if not whdload_dir:
                        # slave was found in root directory
                        pass
                    elif whdload_dir[0] == "/":
                        whdload_dir = whdload_dir[1:]
                    break
            if found_slave:
                break
        if not found_slave:
            raise Exception(
                "Did not find the specified WHDLoad slave {}. "
                "Check the WHDLoad arguments".format(repr(slave_original_name))
            )
        print("[WHDLOAD] Slave directory:", repr(whdload_dir))
        print("[WHDLOAD] Slave arguments:", whdload_args)

        self.copy_whdload_kickstart(
            dest_dir,
            "kick33180.A500",
            ["11f9e62cf299f72184835b7b2a70a16333fc0d88"],
        )
        self.copy_whdload_kickstart(
            dest_dir,
            "kick34005.A500",
            ["891e9a547772fe0c6c19b610baf8bc4ea7fcb785"],
        )
        self.copy_whdload_kickstart(
            dest_dir,
            "kick40068.A1200",
            ["e21545723fe8374e91342617604f1b3d703094f1"],
        )
        self.copy_whdload_kickstart(
            dest_dir,
            "kick40068.A4000",
            ["5fe04842d04a489720f0f4bb0e46948199406f49"],
        )
        whdload.create_prefs_file(
            self.config, os.path.join(s_dir, "WHDLoad.prefs")
        )

        whdload_version = self.config["x_whdload_version"]
        if not whdload_version:
            whdload_version = whdload.default_whdload_version

        for key, value in whdload.binaries[whdload_version].items():
            self.install_whdload_file(key, dest_dir, value)
        for key, value in whdload.support_files.items():
            self.install_whdload_file(key, dest_dir, value)

        if self.config.get("__netplay_game", ""):
            print("[WHDLOAD] Key file is not copied in net play mode")
        else:
            key_file = os.path.join(
                self.fsgs.amiga.get_base_dir(), "WHDLoad.key"
            )
            if os.path.exists(key_file):
                print("found WHDLoad.key at ", key_file)
                shutil.copy(key_file, os.path.join(s_dir, "WHDLoad.key"))
            else:
                print("[WHDLOAD] Key file not found in base dir (FS-UAE dir)")

            # temporary feature, at least until it's possible to set more
            # WHDLoad settings directly in the Launcher
            prefs_file = os.path.join(
                self.fsgs.amiga.get_base_dir(), "WHDLoad.prefs"
            )
            if os.path.exists(prefs_file):
                print("found WHDLoad.prefs at ", prefs_file)
                shutil.copy(prefs_file, os.path.join(s_dir, "WHDLoad.prefs"))
            else:
                print("[WHDLOAD] Key file not found in base dir (FS-UAE dir)")

        if self.config.get("__netplay_game", ""):
            print("[WHDLOAD] WHDLoad base dir is not copied in net play mode")
        else:
            src_dir = self.get_whdload_dir()
            if src_dir and os.path.exists(src_dir):
                print(
                    "[WHDLOAD] WHDLoad base dir exists, copying resources..."
                )
                self.copy_folder_tree(src_dir, dest_dir)

        # icon = self.config.get("__whdload_icon", "")
        icon = ""
        if icon:
            shutil.copy(
                os.path.expanduser("~/kgiconload"),
                os.path.join(dest_dir, "C", "kgiconload"),
            )
            icon_path = os.path.join(dest_dir, icon)
            print("[WHDLOAD] Create icon at ", icon_path)
            create_slave_icon(icon_path, whdload_args)
            self.write_startup_sequence(
                s_dir,
                'cd "{0}"\n'
                "kgiconload {1}\n"
                "uae-configuration SPC_QUIT 1\n".format(
                    whdload_dir, os.path.basename(icon)
                ),
            )
        else:
            self.write_startup_sequence(
                s_dir, whdload_sequence.format(whdload_dir, whdload_args)
            )

    def get_whdload_dir(self):
        path = self.config.get(Option.WHDLOAD_BOOT_DIR)
        return path

    def write_startup_sequence(self, s_dir, command):
        # FIXME: semi-colon is used in WHDLoad CONFIG options...
        command = "\n".join([x.strip() for x in command.split(";")])
        startup_sequence = os.path.join(s_dir, "Startup-Sequence")
        # if True:
        if not os.path.exists(startup_sequence):
            with open(startup_sequence, "wb") as f:
                if "setpatch" in self.hd_requirements:
                    if self.setpatch_installed:
                        f.write(
                            setpatch_found_sequence.replace(
                                "\r\n", "\n"
                            ).encode("ISO-8859-1")
                        )
                    else:
                        f.write(
                            setpatch_not_found_sequence.replace(
                                "\r\n", "\n"
                            ).encode("ISO-8859-1")
                        )
                f.write(command.replace("\r\n", "\n").encode("ISO-8859-1"))

        # The User-Startup file is useful if the user has provided a
        # base WHDLoad directory with an existing startup-sequence
        user_startup = os.path.join(s_dir, "User-Startup")
        with open(user_startup, "ab") as f:
            f.write(command.replace("\r\n", "\n").encode("ISO-8859-1"))

    def install_whdload_file(self, sha1, dest_dir, rel_path):
        abs_path = os.path.join(dest_dir, rel_path)
        name = os.path.basename(rel_path)
        self.on_progress(gettext("Downloading {0}...".format(name)))
        Downloader.install_file_by_sha1(sha1, name, abs_path)

    def copy_setpatch(self, base_dir):
        dest = os.path.join(base_dir, "C")
        if not os.path.exists(dest):
            os.makedirs(dest)
        dest = os.path.join(dest, "SetPatch")
        for checksum in workbench_disks_with_setpatch_39_6:
            path = self.fsgs.file.find_by_sha1(checksum)
            if path:
                print("found WB DISK with SetPatch 39.6 at", path)
                try:
                    input_stream = self.fsgs.file.open(path)
                except Exception:
                    traceback.print_exc()
                else:
                    wb_data = input_stream.read()
                    # archive = Archive(path)
                    # if archive.exists(path):
                    #     f = archive.open(path)
                    #     wb_data = f.read()
                    #     f.close()
                    if self.extract_setpatch_39_6(wb_data, dest):
                        print("SetPatch installed")
                        self.setpatch_installed = True
                        break
                    else:
                        print("WARNING: extract_setpatch_39_6 returned False")
                # else:
                #     print("oops, path does not exist")
        else:
            print("WARNING: did not find SetPatch 39.6")

    @staticmethod
    def extract_setpatch_39_6(wb_data, dest):
        extractor = ADFFileExtractor(wb_data)
        try:
            setpatch_data = extractor.extract_file("C/SetPatch")
        except KeyError:
            return False
        s = hashlib.sha1()
        s.update(setpatch_data)
        print(s.hexdigest())
        # noinspection SpellCheckingInspection
        if s.hexdigest() != "4d4aae988310b07726329e436b2250c0f769ddff":
            return False
        with open(dest, "wb") as f:
            f.write(setpatch_data)
        return True

    def copy_whdload_kickstart(self, base_dir, name, checksums):
        dest = os.path.join(base_dir, "Devs", "Kickstarts")
        if not os.path.exists(dest):
            os.makedirs(dest)
        dest = os.path.join(dest, name)
        for checksum in checksums:
            # print("find kickstart with sha1", checksum)
            path = self.fsgs.file.find_by_sha1(checksum)
            if path:  # and os.path.exists(path):
                print("found kickstart for", name, "at", path)
                archive = Archive(path)
                if archive.exists(path):
                    with open(dest, "wb") as f:
                        ROMManager.decrypt_archive_rom(archive, path, file=f)
                    print(repr(dest))
                    break
                else:
                    stream = self.fsgs.file.open(path)
                    if stream is None:
                        raise Exception("Cannot find kickstart " + repr(path))
                    with open(dest, "wb") as f:
                        f.write(stream.read())

        else:
            print("did not find kickstart for", name)

    def get_state_dir(self):
        return self.game_paths.get_state_dir()

    def init_changes(self):
        print("LaunchHandler.init_changes")
        self.on_progress(gettext("Restoring changes..."))
        self.change_handler.init(
            self.get_state_dir(), ignore=["*.uss", "*.sdf"]
        )

    def update_changes(self):
        print("LaunchHandler.update_changes")
        self.on_progress(gettext("Saving changes..."))
        self.change_handler.update(self.get_state_dir())

    def cleanup(self):
        print("LaunchHandler.cleanup")
        if os.environ.get("FSGS_CLEANUP", "") == "0":
            print("[DRIVER] NOTICE: keeping temp files around...")
            return

        self.on_progress(gettext("Cleaning up..."))
        # self.delete_tree(self.temp_dir)
        shutil.rmtree(self.temp_dir, ignore_errors=True)
        state_dir = self.get_state_dir()
        try:
            # this will only succeed if the directory is empty -we don't
            # want to leave unnecessary empty state directories
            os.rmdir(state_dir)
            print("removed", repr(state_dir))
            # also try to remove the parent (letter dir)
            os.rmdir(os.path.dirname(state_dir))
            print("removed", repr(os.path.dirname(state_dir)))
        except OSError:
            # could not delete directories - ok - probably has content
            pass

    def prepare_theme(self):
        # path = self.game_paths.get_theme_path()
        # if path:
        #     self.config["theme"] = path
        pass

    def prepare_extra_settings(self):
        prefix = self.config.get("screenshots_output_prefix", "")
        if prefix:
            return
        # name = self.config.get("floppy_drive_0", "")
        # if not name:
        #     name = self.config.get("hard_drive_0", "")
        # if not name:
        #     name = self.config.get("cdrom_drive_0", "")
        # if not name:
        #     name = self.config.get("floppy_image_0", "")
        name = self.config_name
        if not name:
            name = "fs-uae"
        name, variant = GameNameUtil.extract_names(name)
        name = GameNameUtil.create_cmpname(name)
        self.config["screenshots_output_prefix"] = name

    def create_config(self):
        config = ConfigWriter(self.config).create_fsuae_config()
        return config

    def write_config(self, f):
        config_lines = self.create_config()
        for line in config_lines:
            f.write(line)
            f.write("\n")

    def write_config_to_file(self, path):
        with open(path, "wb") as f:
            self.write_config(f)

    def run(self):
        print("LaunchHandler.run")
        # self.on_progress(gettext("Starting FS-UAE..."))
        # current_task.set_progress(gettext("Starting FS-UAE..."))
        current_task.set_progress("__run__")
        config = self.create_config()
        if self.use_relative_paths:
            process, config_file = FSUAE.start_with_config(
                config, cwd=self.temp_dir
            )
        else:
            process, config_file = FSUAE.start_with_config(config)
        pid_file = self.fsgc.settings[Option.EMULATOR_PID_FILE]
        GameDriver.write_emulator_pid_file(pid_file, process)
        process.wait()
        GameDriver.remove_emulator_pid_file(pid_file)
        print("LaunchHandler.start is done")
        if os.environ.get("FSGS_CLEANUP", "") == "0":
            print("Not removing", config_file)
        else:
            print("removing", config_file)
            try:
                os.remove(config_file)
            except Exception:
                print("could not remove config file", config_file)

    def unpack_archive(self, path, destination):
        print("unpack", path, "to", destination)
        archive = Archive(path)
        print(archive)
        print(archive.get_handler())
        for entry in archive.list_files():
            if self.stop_flag:
                return

            print(entry)
            n = entry[len(path) + 2 :]
            amiga_rel_path = amiga_path_to_host_path(n)

            out_path = os.path.join(destination, amiga_rel_path)
            print("out path", out_path)

            if entry.endswith("/"):
                os.makedirs(out_path)
            else:
                if not os.path.exists(os.path.dirname(out_path)):
                    os.makedirs(os.path.dirname(out_path))
                f = archive.open(entry)
                with open(out_path, "wb") as out_f:
                    while True:
                        data = f.read(65536)
                        if not data:
                            break
                        out_f.write(data)
                # FIXME: Extract real timestamps from archive
                # FIXME: Real metadata from archive
                # noinspection SpellCheckingInspection
                metadata = [
                    "----rwed",
                    " ",
                    "2000-01-01 00:00:00.00",
                    " ",
                    "",
                    "\n",
                ]
                info = archive.getinfo(entry)
                if info.comment:
                    # print(info.comment)
                    # raise Exception("gnit")
                    metadata[4] = self.encode_file_comment(info.comment)
                with open(out_path + ".uaem", "wb") as out_file:
                    out_file.write("".join(metadata).encode("UTF-8"))

    def copy_folder_tree(self, source_path, dest_path, overwrite=False):
        for item in os.listdir(source_path):
            if self.stop_flag:
                return
            if item[0] == ".":
                continue
            item_path = os.path.join(source_path, item)
            dest_item_path = os.path.join(dest_path, item)
            if os.path.isdir(item_path):
                if not os.path.exists(dest_item_path):
                    os.makedirs(dest_item_path)
                self.copy_folder_tree(item_path, dest_item_path)
                if self.stop_flag:
                    return
            else:
                if overwrite or not os.path.exists(dest_item_path):
                    print("copy", repr(item_path), "to", repr(dest_item_path))
                    shutil.copy(item_path, dest_item_path)


def amiga_path_to_host_path(amiga_rel_path):
    # amiga_rel_path = amiga_path
    print("amiga_rel_path", amiga_rel_path)
    amiga_rel_parts = amiga_rel_path.split("/")
    for i, part in enumerate(amiga_rel_parts):
        # part can be blank if amiga_rel_parts is a directory
        # (ending with /)
        if part:
            amiga_rel_parts[i] = amiga_filename_to_host_filename(part)
    amiga_rel_path = "/".join(amiga_rel_parts)
    return amiga_rel_path


def amiga_filename_to_host_filename(amiga_name, ascii_only=False):
    """
    Converted from FS-UAE C code (src/od-fs/fsdb-host.py)
    @author: TheCyberDruid
    """
    length = len(amiga_name)

    replace_1 = -1
    replace_2 = -1

    check = amiga_name[:3].upper()
    dot_pos = -1
    if check in ["AUX", "CON", "PRN", "NUL"]:
        dot_pos = 4
    elif check in ["LPT", "COM"] and length >= 4 and amiga_name[3].isdigit():
        dot_pos = 5
    if dot_pos > -1 and (
        length == (dot_pos - 1)
        or (length > dot_pos and amiga_name[dot_pos] == ".")
    ):
        replace_1 = 2

    if amiga_name[-1] == "." or amiga_name[-1] == " ":
        replace_2 = length - 1

    i = 0
    filename = ""
    for c in amiga_name:
        x = ord(c)
        replace = False
        if i == replace_1:
            replace = True
        elif i == replace_2:
            replace = True
        elif x < 32:
            replace = True
        elif ascii_only and x > 127:
            replace = True

        if not replace:
            for evil in EVIL_CHARS:
                if evil == c:
                    replace = True
                    break
        if (i == length - 1) and amiga_name[-5:] == ".uaem":
            replace = True

        if replace:
            filename += "%" + "%02x" % ord(c)
        else:
            filename += c

        i += 1

    return filename


EVIL_CHARS = '%\\*?"/|<>'

system_configuration = (
    b"\x08\x00\x00\x05\x00\x00\x00\x00\x00\x00\xc3"
    b"P\x00\x00\x00\x00\x00\t'\xc0\x00\x00\x00\x01\x00\x00N \x00\x00\x00\x00"
    b"\xc0\x00@\x00p\x00\xb0\x00<\x00L\x00?\x00C\x00\x1f\xc0 \xc0\x1f\xc0 \x00"
    b"\x0f\x00\x11\x00\r\x80\x12\x80\x04\xc0\t@\x04`\x08\xa0\x00 \x00@\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\xff\x00\x0eD\x00\x00\x0e\xec\x00\x01\n\xaa\x00\x00\x0f"
    b"\xff\x06\x8b\x00\x00\x00\x81\x00,\x00\x00\x00\x00generic\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x05\x00K\x00\x00\x00\x00\x00\x00\x00\x07"
    b"\x00 \x00B\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00"
    b"\x00"
)

# noinspection SpellCheckingInspection
workbench_disks_with_setpatch_39_6 = [
    # Workbench v3.0 rev 39.29 (1992)(Commodore)(A1200-A4000)(M10)
    # (Disk 1 of 6)(Install).adf
    "ba24b4172339b9198e4f724a6804d0c6eb5e394b",
    # Workbench v3.0 rev 39.29 (1992)(Commodore)(A1200-A4000)(M10)
    # (Disk 1 of 6)(Install)[a].adf
    "c0781dece2486b54e15ce54a9b24dec6d9429421",
    # Workbench v3.0 rev 39.29 (1992)(Commodore)(A1200-A4000)(M10)
    # (Disk 1 of 6)(Install)[m drive definitions].adf
    "7eeb2511ce34f8d3f09efe82b290bddeb899d237",
    # Workbench v3.0 rev 39.29 (1992)(Commodore)(A1200-A4000)(M10)
    # (Disk 1 of 6)(Install)[m2].adf
    "7271d7db4472e10fbe4b266278e16f03336c14e3",
    # Workbench v3.0 rev 39.29 (1992)(Commodore)(A1200-A4000)(M10)
    # (Disk 1 of 6)(Install)[m3].adf
    "92c2f33bb73e1bdee5d9a0dc0f5b09a15524f684",
    # Workbench v3.0 rev 39.29 (1992)(Commodore)(A1200-A4000)(M10)
    # (Disk 2 of 6)(Workbench)[!].adf
    "e663c92a9c88fa38d02bbb299bea8ce70c56b417",
    # Workbench v3.0 rev 39.29 (1992)(Commodore)(A1200-A4000)(M10)
    # (Disk 2 of 6)(Workbench)[a2].adf
    "65ab988e597b456ac40320f88a502fc016d590aa",
    # Workbench v3.0 rev 39.29 (1992)(Commodore)(A1200-A4000)(M10)
    # (Disk 2 of 6)(Workbench)[a].adf
    "9496daa66e6b2f4ddde4fa2580bb3983a25e3cd2",
    # Workbench v3.0 rev 39.29 (1992)(Commodore)(A1200-A4000)(M10)
    # (Disk 2 of 6)(Workbench)[m2].adf
    "cf2f24cf5f5065479476a38ec8f1016b1f746884",
    # Workbench v3.0 rev 39.29 (1992)(Commodore)(A1200-A4000)(M10)
    # (Disk 2 of 6)(Workbench)[m3].adf
    "0e7f30223af254df0e2b91ea409f35c56d6164a6",
    # Workbench v3.0 rev 39.29 (1992)(Commodore)(A1200-A4000)(M10)
    # (Disk 2 of 6)(Workbench)[m4].adf
    "08c4afde7a67e6aaee1f07af96e95e9bed897947",
    # amiga-os-300-workbench.adf
    # Workbench v3.0 rev 39.29 (1992)(Commodore)(A1200-A4000)(M10)
    # (Disk 2 of 6)(Workbench)[m5].adf
    "4f4770caae5950eca4a2720e0424df052ced6a32",
    # Workbench v3.0 rev 39.29 (1992)(Commodore)(A1200-A4000)(M10)
    # (Disk 2 of 6)(Workbench)[m].adf
    "53086c3e44ec2d34e60ab65af71fb11941f4e0af",
]

setpatch_found_sequence = """
C:SetPatch
"""

setpatch_not_found_sequence = """
echo "Warning: SetPatch (39.6) not found."
echo "Make sure a WB 3.0 disk is scanned in FS-UAE Launcher"
echo "and the file will automatically be copied from the disk."
EndIF
"""

whdload_sequence = """
cd "{0}"
WHDLoad {1}
uae-configuration SPC_QUIT 1
"""

# noinspection SpellCheckingInspection
base_icon = (
    b"\xe3\x10\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x006\x00"
    b"\x17\x00\x04\x00\x01\x00\x01\x00\x02\x1a\x98\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
    b"\x00\x04\x00\x00\t\x01l\x00\x01\xa5\x9c\x80\x00\x00\x00\x80\x00"
    b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00"
    b"\x00\x00\x006\x00\x16\x00\x02\x00\x08\xce \x03\x00\x00\x00\x00\x00"
    b"\x00\x00\x00\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x0c\x00"
    b"\x00\x00\x00\x00\x00\x00\x0c\x00\x00\x00\x00\x00\x00\x00\x0c\x00"
    b"\x00\x00\x00\x00\x00\x00\x0c\x00\x00\x00\x00\x00\x00\x00\x0c\x00"
    b"\x03\xf0\x0f\xff\xe0\x00\x0c\x00\x02\x080\x00\x1c\x00\x0c\x00\x02"
    b"\x07\xc0\x00\x03\x80\x0c\x00\x02\x00\x00\x00\x00`\x0c\x00\x02\x00"
    b"\x00\x00\x00\x10\x0c\x00\x02\x00\x00\x00\x00\x08\x0c\x00\x02\x07"
    b"\xc0\x00\x1f\xc4\x0c\x00\x02\x08 \x00 2\x0c\x00\x03\xf0\x18\x00"
    b"\xc0\r\x0c\x00\x00\x00\x06\x03\x00\x03\x0c\x00\x00\x00\x02\x02"
    b"\x00\x00\x0c\x00\x00\x00\x02\x02\x00\x00\x0c\x00\x00\x00\x02\x02"
    b"\x00\x00\x0c\x00\x00\x00\x03\xfe\x00\x00\x0c\x00\x00\x00\x00\x00"
    b"\x00\x00\x0c\x00\x7f\xff\xff\xff\xff\xff\xfc\x00\xff\xff\xff\xff"
    b"\xff\xff\xf8\x00\xd5UUUUUP\x00\xd5UUUUUP\x00\xd5UUUUUP\x00\xd5UUU"
    b"UUP\x00\xd5UUUUUP\x00\xd4\x05P\x00\x15UP\x00\xd4\x05@\x00\x01UP"
    b"\x00\xd4\x00\x00\x00\x00UP\x00\xd4\x00\x00\x00\x00\x15P\x00\xd4"
    b"\x00\x00\x00\x00\x05P\x00\xd4\x00\x00\x00\x00\x05P\x00\xd4\x00"
    b"\x00\x00\x00\x01P\x00\xd4\x05@\x00\x15AP\x00\xd4\x05@\x00\x15PP"
    b"\x00\xd5UP\x00UTP\x00\xd5UT\x01UUP\x00\xd5UT\x01UUP\x00\xd5UT"
    b"\x01UUP\x00\xd5UT\x01UUP\x00\xd5UUUUUP\x00\x80\x00\x00\x00\x00"
    b"\x00\x00\x00"
)


def char(v):
    return chr(v)


def write_number(f, n):
    f.write(char((n & 0xFF000000) >> 3))
    f.write(char((n & 0x00FF0000) >> 2))
    f.write(char((n & 0x0000FF00) >> 1))
    f.write(char((n & 0x000000FF) >> 0))


def write_string(f, s):
    write_number(f, len(s) + 1)
    f.write(s)
    f.write("\0")


def create_slave_icon(path, whdload_args):
    default_tool = "DH0:/C/WHDLoad"
    # FIXME: handle "" around slave name?
    # args = whdload_args.split(" ")
    tool_types = whdload_args.split(" ")
    tool_types[0] = "SLAVE=" + tool_types[0]

    with open(path, "wb") as f:
        f.write(base_icon)
        write_string(f, default_tool)
        write_number(f, (len(tool_types) + 1) * 4)
        for tool_type in tool_types:
            write_string(f, tool_type)
        f.close()


def is_sha1(s):
    if len(s) != 40:
        return False
    for c in s:
        if c not in "0123456789abcdef":
            return False
    return True



import json
import time
from gzip import GzipFile
from io import StringIO
from urllib.request import Request

import requests

from fsgs.network import is_http_url, openretro_url_prefix
from fsgs.res import gettext

# FIXME: Overlap in code/functionality with GameDatabaseSynchronizer


class SynchronizerBase(object):
    def __init__(self, context, on_status=None, stop_check=None):
        self.context = context
        self.on_status = on_status
        self._stop_check = stop_check

    @staticmethod
    def bytes_to_int(m):
        return m[0] << 24 | m[1] << 16 | m[2] << 8 | m[3]

    def stop_check(self):
        if self._stop_check:
            return self._stop_check()

    def set_status(self, title, status=""):
        if self.on_status:
            self.on_status((title, status))

    def url_prefix(self):
        return openretro_url_prefix()

    def auth(self):
        return self.context.username, self.context.password

    def url(self, url):
        if not is_http_url(url):
            url = "{0}{1}".format(self.url_prefix(), url)
        return url

    def fetch_json_attempt(self, url):
        print("[HTTP] {}".format(self.url(url)))
        r = requests.get(self.url(url), auth=self.auth())
        r.raise_for_status()
        return r.json()

    def fetch_data_attempt(self, url, accept_gzip_encoding=False):
        print("[HTTP] {}".format(self.url(url)))
        r = requests.get(self.url(url), auth=self.auth())
        r.raise_for_status()
        return r.content

    def fetch_json(self, url):
        for i in range(20):
            try:
                return self.fetch_json_attempt(url)
            except Exception as e:
                print(e)
                sleep_time = 2.0 + i * 0.3
                # FIXME: change second {0} to {1}
                self.set_status(
                    gettext(
                        "Download failed (attempt {0}) - retrying in {0} "
                        "seconds"
                    ).format(i + 1, int(sleep_time))
                )
                for _ in range(int(sleep_time) * 10):
                    time.sleep(0.1)
                    if self.stop_check():
                        return
                self.set_status(
                    gettext("Retrying last operation (attempt {0})").format(
                        i + 1
                    )
                )
        return self.fetch_json_attempt(url)

    def fetch_data(self, url):
        for i in range(10):
            try:
                return self.fetch_data_attempt(url)
            except Exception as e:
                print(e)
                sleep_time = 2.0 + i * 0.3
                # FIXME: change second {0} to {1}
                self.set_status(
                    gettext(
                        "Download failed (attempt {0}) - retrying in {0} "
                        "seconds"
                    ).format(i + 1, int(sleep_time))
                )
                for _ in range(int(sleep_time) * 10):
                    time.sleep(0.1)
                    if self.stop_check():
                        return
                self.set_status(
                    gettext("Retrying last operation (attempt {0})").format(
                        i + 1
                    )
                )
        return self.fetch_data_attempt(url)



from __future__ import absolute_import
from __future__ import print_function

from .block.BootBlock import BootBlock
from .block.RootBlock import RootBlock
from .ADFSVolDir import ADFSVolDir
from .ADFSBitmap import ADFSBitmap
from .FileName import FileName
from .RootMetaInfo import RootMetaInfo
from .FSError import *
from .FSString import FSString
from .TimeStamp import TimeStamp
from . import DosType
import amitools.util.ByteSize as ByteSize

class ADFSVolume:
  
  def __init__(self, blkdev):
    self.blkdev = blkdev
    
    self.boot = None
    self.root = None
    self.root_dir = None
    self.bitmap = None
    
    self.valid = False
    self.is_ffs = None
    self.is_intl = None
    self.is_dircache = None
    self.is_longname = None
    self.name = None
    self.meta_info = None
    
  def open(self):
    # read boot block
    self.boot = BootBlock(self.blkdev)
    self.boot.read()
    # valid root block?
    if self.boot.valid:
      # get fs flags
      dos_type = self.boot.dos_type
      self.is_ffs = DosType.is_ffs(dos_type)
      self.is_intl = DosType.is_intl(dos_type)
      self.is_dircache = DosType.is_dircache(dos_type)
      self.is_longname = DosType.is_longname(dos_type)
      # read root 
      self.root = RootBlock(self.blkdev, self.boot.calc_root_blk)
      self.root.read()
      if self.root.valid:
        self.name = FSString(self.root.name)
        # build meta info
        self.meta_info = RootMetaInfo( self.root.create_ts, self.root.disk_ts, self.root.mod_ts )
        # create root dir
        self.root_dir = ADFSVolDir(self, self.root)
        self.root_dir.read()
        # create bitmap
        self.bitmap = ADFSBitmap(self.root)
        self.bitmap.read()
        self.valid = True
      else:
        raise FSError(INVALID_ROOT_BLOCK, block=self.root)
    else:
      raise FSError(INVALID_BOOT_BLOCK, block=self.boot)
  
  def create(self, name, meta_info=None, dos_type=None, boot_code=None, is_ffs=False, is_intl=False, is_dircache=False, is_longname=False):
    # determine dos_type
    if dos_type == None:
      dos_type = DosType.DOS0
      if is_longname:
        dos_type = DosType.DOS6
      elif is_dircache:
        dos_type |= DosType.DOS_MASK_DIRCACHE
      elif is_intl:
        dos_type |= DosType.DOS_MASK_INTL
      if is_ffs:
        dos_type |= DosType.DOS_MASK_FFS
    # update flags
    self.is_ffs = DosType.is_ffs(dos_type)
    self.is_intl = DosType.is_intl(dos_type)
    self.is_dircache = DosType.is_dircache(dos_type)
    self.is_longname = DosType.is_longname(dos_type)
    # convert and check volume name
    if not isinstance(name, FSString):
      raise ValueError("create's name must be a FSString")
    fn = FileName(name, is_intl=self.is_intl, is_longname=False) # Volumes don't support long names
    if not fn.is_valid():
      raise FSError(INVALID_VOLUME_NAME, file_name=name, node=self)
    # create a boot block
    self.boot = BootBlock(self.blkdev)
    self.boot.create(dos_type=dos_type, boot_code=boot_code)
    self.boot.write()
    # create a root block
    self.root = RootBlock(self.blkdev, self.boot.calc_root_blk)
    if meta_info == None:
      meta_info = RootMetaInfo()
      meta_info.set_current_as_create_time()
      meta_info.set_current_as_mod_time()
      meta_info.set_current_as_disk_time()
    create_ts = meta_info.get_create_ts()
    disk_ts = meta_info.get_disk_ts()
    mod_ts = meta_info.get_mod_ts()
    self.meta_info = meta_info
    self.root.create(name.get_ami_str(), create_ts, disk_ts, mod_ts, fstype=dos_type)
    self.name = name
    # create bitmap
    self.bitmap = ADFSBitmap(self.root)
    self.bitmap.create()
    self.bitmap.write() # writes root block, too
    # create empty root dir
    self.root_dir = ADFSVolDir(self, self.root)
    self.root_dir.read()
    # all ok
    self.valid = True
  
  def close(self):
    pass

  def get_info(self):
    """return an array of strings with information on the volume"""
    res = []
    total = self.get_total_blocks()
    free  = self.get_free_blocks()
    used  = total - free
    bb = self.blkdev.block_bytes
    btotal = total * bb
    bfree  = free * bb
    bused  = used * bb
    prc_free = 10000 * free / total
    prc_used = 10000 - prc_free
    res.append("total:  %10d  %s  %12d" % (total, ByteSize.to_byte_size_str(btotal), btotal))
    res.append("used:   %10d  %s  %12d  %5.2f%%" % (used, ByteSize.to_byte_size_str(bused), bused, prc_used / 100.0))
    res.append("free:   %10d  %s  %12d  %5.2f%%" % (free, ByteSize.to_byte_size_str(bfree), bfree, prc_free / 100.0))
    return res

  # ----- Path Queries -----

  def get_path_name(self, path_name, allow_file=True, allow_dir=True):
    """get node for given path"""
    # make sure path name is a FSString
    if not isinstance(path_name, FSString):
      raise ValueError("get_path_name's path must be a FSString")
    # create and check file name
    fn = FileName(path_name, is_intl=self.is_intl, is_longname=self.is_longname)
    if not fn.is_valid():
      raise FSError(INVALID_FILE_NAME, file_name=path_name, node=self)      
    # find node
    if fn.is_root_path_alias():
      # its the root node
      return self.root_dir
    else:
      # find a sub node
      path = fn.split_path()
      return self.root_dir.get_path(path, allow_file, allow_dir)

  def get_dir_path_name(self, path_name):
    """get node for given path and ensure its a directory"""
    return self.get_path_name(path_name, allow_file=False)
  
  def get_file_path_name(self, path_name):
    """get node for given path and ensure its a file"""
    return self.get_path_name(path_name, allow_dir=False)
  
  def get_create_path_name(self, path_name, suggest_name=None):
    """get a parent node and path name for creation
       return: parent_node_or_none, file_name_or_none
    """
    # make sure input is correct
    if not isinstance(path_name, FSString):
      raise ValueError("get_create_path_name's path_name must be a FSString")
    if suggest_name != None and not isinstance(suggest_name, FSString):
      raise ValueError("get_create_path_name's suggest_name must be a FSString")
    # is root path?
    fn = FileName(path_name, is_intl=self.is_intl, is_longname=self.is_longname)
    if not fn.is_valid():
      raise FSError(INVALID_FILE_NAME, file_name=path_name, node=self)      
    # find node
    if fn.is_root_path_alias():
      return self.root_dir, suggest_name
    else:
      # try to get path_name as a directory
      node = self.get_dir_path_name(path_name)
      if node != None:
        return node, suggest_name
      else:
        # split into dir and file name
        dn, fn = fn.get_dir_and_base_name()
        if dn != None:
          # has a directory -> try to fetch it
          node = self.get_dir_path_name(dn)
        else:
          # no dir -> assume root dir
          node = self.root_dir
        if fn != None:
          # take given name
          return node, fn
        else:
          # use suggested name
          return node, suggest_name
    
  # ----- convenience API -----
  
  def get_volume_name(self):
    return self.name
  
  def get_root_dir(self):
    return self.root_dir
  
  def get_dos_type(self):
    return self.boot.dos_type
    
  def get_boot_code(self):
    return self.boot.boot_code
  
  def get_free_blocks(self):
    return self.bitmap.get_num_free()
    
  def get_used_blocks(self):
    free = self.bitmap.get_num_free()
    total = self.blkdev.num_blocks
    return total - free
    
  def get_total_blocks(self):
    return self.blkdev.num_blocks
  
  def get_meta_info(self):
    return self.meta_info

  def update_disk_time(self):
    mi = RootMetaInfo()
    mi.set_current_as_disk_time()
    self.change_meta_info(mi)
    
  def change_meta_info(self, meta_info):
    if self.root != None and self.root.valid:
      dirty = False
      # update create_ts
      create_ts = meta_info.get_create_ts()
      if create_ts != None:
        self.root.create_ts = meta_info.get_create_ts()
        dirty = True
      # update disk_ts
      disk_ts = meta_info.get_disk_ts()
      if disk_ts != None:
        self.root.disk_ts = disk_ts
        dirty = True
      # update mod_ts
      mod_ts = meta_info.get_mod_ts()
      if mod_ts != None:
        self.root.mod_ts = mod_ts
        dirty = True
      # update if something changed
      if dirty:
        self.root.write()
        self.meta_info = RootMetaInfo( self.root.create_ts, self.root.disk_ts, self.root.mod_ts )
      return True
    else:
      return False
      
  def change_create_ts(self, create_ts):
    return self.change_meta_info(RootMetaInfo(create_ts=create_ts))
    
  def change_disk_ts(self, disk_ts):
    return self.change_meta_info(RootMetaInfo(disk_ts=disk_ts))
    
  def change_mod_ts(self, mod_ts):
    return self.change_meta_info(RootMetaInfo(mod_ts=mod_ts))
  
  def change_create_ts_by_string(self, create_ts_str):
    t = TimeStamp()
    t.parse(create_ts_str)
    return self.change_meta_info(RootMetaInfo(create_ts=t))

  def change_disk_ts_by_string(self, disk_ts_str):
    t = TimeStamp()
    t.parse(disk_ts_str)
    return self.change_meta_info(RootMetaInfo(disk_ts=t))
    
  def change_mod_ts_by_string(self, mod_ts_str):
    t = TimeStamp()
    t.parse(mod_ts_str)
    return self.change_meta_info(RootMetaInfo(mod_ts=t))    

  def relabel(self, name):
    """Relabel the volume"""
    # make sure its a FSString
    if not isinstance(name, FSString):
      raise ValueError("relabel's name must be a FSString")
    # validate file name
    fn = FileName(name, is_intl=self.is_intl, is_longname=False)
    if not fn.is_valid():
      raise FSError(INVALID_VOLUME_NAME, file_name=name, node=self)
    # update root block
    ami_name = name.get_ami_str()
    self.root.name = ami_name
    self.root.write()
    # store internally
    self.name = name
    self.root_dir.name = name

  def create_dir(self, ami_path):
    """Create a new directory"""
    # make sure its a FSString
    if not isinstance(ami_path, FSString):
      raise ValueError("create_dir's ami_path must be a FSString")
    # check file path
    fn = FileName(ami_path, is_intl=self.is_intl, is_longname=self.is_longname)
    if not fn.is_valid():
      raise FSError(INVALID_FILE_NAME, file_name=ami_path)
    # split into dir and base name
    dir_name, base_name = fn.get_dir_and_base_name()
    if base_name == None:
      raise FSError(INVALID_FILE_NAME, file_name=ami_path)
    # find parent of dir
    if dir_name == None:
      node = self.root_dir
    else:
      # no parent dir found
      node = self.get_dir_path_name(dir_name)
      if node == None:
        raise FSError(INVALID_PARENT_DIRECTORY, file_name=ami_path, extra="not found: "+dir_name)
    node.create_dir(base_name)
    
  def write_file(self, data, ami_path, suggest_name=None, cache=False):
    """Write given data as a file"""
    # get parent node and file_name
    parent_node, file_name = self.get_create_path_name(ami_path, suggest_name)
    if parent_node == None:
      raise FSError(INVALID_PARENT_DIRECTORY, file_name=ami_path)
    if file_name == None:
      raise FSError(INVALID_FILE_NAME, file_name=file_name)
    # create file
    node = parent_node.create_file(file_name, data)
    if not cache:
      node.flush()
  
  def read_file(self, ami_path, cache=False):
    """Read a file and return data"""
    # get node of file
    node = self.get_file_path_name(ami_path)
    if node == None:
      raise FSError(FILE_NOT_FOUND, file_name=ami_path)
    data = node.get_file_data()
    if not cache:
      node.flush()
    return data
  
  def delete(self, ami_path, wipe=False, all=False):
    """Delete a file or directory at given path"""
    node = self.get_path_name(ami_path)
    if node == None:
      raise FSError(FILE_NOT_FOUND, file_name=ami_path)
    node.delete(wipe=wipe, all=all)



import weakref
from fsui.qt import QDialog, Signal
from fsui.qt.helpers import QParent


class LegacyDialog(QDialog):
    closed = Signal()

    def __init__(self, parent=None, title=""):
        QDialog.__init__(self, QParent(parent))
        self._window = weakref.ref(self)

        self.layout = None
        self.setWindowTitle(title)

        # self.container = wx.Panel(self)
        # self.container.get_window = self.get_window
        # self.Bind(wx.EVT_SIZE, self.__resize_event)
        # self.Bind(wx.EVT_WINDOW_DESTROY, self.__destroy_event)
        # self.Bind(wx.EVT_CLOSE, self.__close_event)
        self.destroy_listeners = []
        self.close_listeners = []

    def get_parent(self):
        return None

    def closeEvent(self, event):
        print("Dialog.closeEvent")
        self.closed.emit()
        self.on_close()
        for function in self.close_listeners:
            print(function)
            function()
        event.accept()
        # remove close listeners so they will not keep the object alive
        self.close_listeners = []

    def add_close_listener(self, function):
        self.close_listeners.append(function)

    def get_window(self):
        return self

    def get_container(self):
        return self

    # def show(self):
    #     self.Show()

    # def close(self):
    #     self.Close()

    def show_modal(self):
        # self.setModal(True)
        # return self.showModal()
        return self.exec()

    def end_modal(self, value):
        # self.EndModal(value)
        self.done(value)

    def center_on_parent(self):
        real_parent = self.parent()
        if real_parent:
            pp = real_parent.x(), real_parent.y()
            ps = real_parent.width(), real_parent.height()
            ss = self.get_size()
            self.move(
                pp[0] + (ps[0] - ss[0]) // 2, pp[1] + (ps[1] - ss[1]) // 2
            )
        # elif self.default_center:
        #     x, y = self.default_center
        #     ss = self.get_size()
        #     self.move(x - ss[0] // 2, y - ss[1] // 2,)

    # def destroy(self):
    #     #self.Destroy()
    #     print("FIXME: Dialog.destroy does nothing")

    def set_title(self, title):
        self.setWindowTitle(title)

    def set_size(self, size):
        # self.SetClientSize(size)
        # print("FIXME:\n\nDialog.set_size")
        self.resize(size[0], size[1])

    def on_create(self):
        pass

    def on_close(self):
        pass

    def on_destroy(self):
        pass

    def __destroy_event(self, event):
        self.on_destroy()

    def __close_event(self, event):
        print("__close_event")
        for function in self.close_listeners:
            function()
        self.on_close()
        self.Destroy()

    def showEvent(self, event):
        self.on_resize()

    def get_size(self):
        return self.width(), self.height()

    def resizeEvent(self, event):
        print("resized..")
        self.on_resize()

    def on_resize(self):
        if self.layout:
            self.layout.set_size(self.get_size())
            self.layout.update()

    def raise_and_activate(self):
        self.raise_()
        self.activateWindow()



from fsgs.platforms.amiga import AmigaPlatformHandler


class CD32PlatformHandler(AmigaPlatformHandler):
    PLATFORM_NAME = "CD32"

    def __init__(self):
        AmigaPlatformHandler.__init__(self)



TOP_ITEM_NOBORDER = 0
TOP_ITEM_ARROW = 1
TOP_ITEM_LEFT = 2
TOP_ITEM_RIGHT = 3

ROW_NAME = 1
ROW_PLATFORM = 3

TOP_WIDTH = 64

COVER_POS = 0.75
# TOP_HEIGHT = 1.0 / 8.0
# TOP_HEIGHT = 60

# actual/visible height is 60, height incl shadow is 90
# - not using shadow anymore
# TOP_HEIGHT = 90
TOP_HEIGHT = 60
TOP_HEIGHT_VISIBLE = 60



import weakref

# noinspection PyUnresolvedReferences
from fsui.qt import Signal


class SignalInhibitor:
    def __init__(self):
        self._inhibited = False

    def __enter__(self):
        self._inhibited = True

    def __exit__(self, exc_type, exc_value, traceback):
        self._inhibited = False

    def __bool__(self):
        return self._inhibited


class SignalWrapper:
    def __init__(self, parent, name):
        self.parent = weakref.ref(parent)
        self.name = name
        self.inhibit = SignalInhibitor()

    def connect(self, obj):
        getattr(self.parent(), self.name + "_signal").connect(obj)

    def emit(self, *args, **kwargs):
        getattr(self.parent(), self.name + "_signal").emit(*args, **kwargs)



import fsui

from ...devicemanager import DeviceManager
from ...i18n import gettext
from ..IconButton import IconButton
from .InputSelector import InputSelector


class InputGroup(fsui.Group):
    def __init__(
        self,
        parent,
        autofire_button=True,
        refresh_button=False,
        parallel_ports=False,
        custom_ports=False,
    ):
        fsui.Group.__init__(self, parent)
        self.layout = fsui.VerticalLayout()

        if parallel_ports:
            heading = gettext("Parallel Port Joysticks")
        elif custom_ports:
            heading = gettext("Custom Joystick Port")
        else:
            heading = gettext("Joystick & Mouse Port")
            if False:
                # Keeping the old string here to keep the translations alive
                gettext("Joystick Ports")

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)

        heading_label = fsui.HeadingLabel(self, heading)
        hori_layout.add(heading_label, margin=10)
        hori_layout.add_spacer(0, expand=True)

        if refresh_button:
            self.refresh_button = IconButton(self, "refresh_button.png")
            self.refresh_button.set_tooltip(
                gettext("Refresh List of Connected Joystick Devices")
            )
            self.refresh_button.activated.connect(self.on_refresh_button)
            hori_layout.add(self.refresh_button, margin_right=10)

        self.layout.add_spacer(0)

        self.selectors = []
        offset = 0
        count = 2
        if parallel_ports:
            offset = 2
        elif custom_ports:
            offset = 4
            count = 1

        input_ports = [1, 0, 2, 3, 4]
        for i in input_ports[offset : offset + count]:
            # self.layout.add_spacer(10)
            selector = InputSelector(self, i, autofire_button=autofire_button)
            self.layout.add(selector, fill=True, margin=10)

    # noinspection PyMethodMayBeStatic
    def on_refresh_button(self):
        DeviceManager.refresh()



import traceback

import fsui
from fsgs.context import fsgs
from fsgs.Database import Database
from launcher.ui.ConfigGroup import ConfigGroup

from ..launcher_config import LauncherConfig
from ..launcher_settings import LauncherSettings
from ..launcher_signal import LauncherSignal


class LastVariants(object):
    def __init__(self):
        self.cache = {}
        LauncherSignal.add_listener("quit", self)

    def on_quit_signal(self):
        database = Database.get_instance()
        for key, value in self.cache.items():
            database.set_last_game_variant(key, value)
        database.commit()


# class VariantsBrowser(fsui.VerticalItemView):
class VariantsBrowser(fsui.ItemChoice):
    @staticmethod
    def use_horizontal_layout():
        # return fsui.get_screen_size()[0] > 1024
        return False

    def __init__(self, parent):
        # fsui.VerticalItemView.__init__(self, parent)
        fsui.ItemChoice.__init__(self, parent)

        self.parent_uuid = ""
        self.items = []  # type: list [dict]
        # self.last_variants = LastVariants()

        self.icon = fsui.Image("launcher:res/fsuae_config_16.png")
        self.adf_icon = fsui.Image("launcher:res/adf_game_16.png")
        self.ipf_icon = fsui.Image("launcher:res/ipf_game_16.png")
        self.cd_icon = fsui.Image("launcher:res/cd_game_16.png")
        self.hd_icon = fsui.Image("launcher:res/hd_game_16.png")
        # self.missing_icon = fsui.Image(
        #     "launcher:res/missing_game_16.png")
        self.missing_icon = fsui.Image("launcher:res/16x16/delete_grey.png")
        self.missing_color = fsui.Color(0xA8, 0xA8, 0xA8)

        self.blank_icon = fsui.Image("launcher:res/16x16/blank.png")
        self.bullet_icon = fsui.Image("launcher:res/16x16/bullet.png")

        self.manual_download_icon = fsui.Image(
            "launcher:res/16x16/arrow_down_yellow.png"
        )
        self.auto_download_icon = fsui.Image(
            "launcher:res/16x16/arrow_down_green.png"
        )

        self.up_icon = fsui.Image("launcher:res/16x16/thumb_up_mod.png")
        self.down_icon = fsui.Image("launcher:res/16x16/thumb_down_mod.png")
        self.fav_icon = fsui.Image("launcher:res/rating_fav_16.png")

        LauncherSettings.add_listener(self)
        self.on_setting("parent_uuid", LauncherSettings.get("parent_uuid"))

    def on_destroy(self):
        LauncherSettings.remove_listener(self)
        # Signal.remove_listener("quit", self)

    def on_select_item(self, index):
        if index is None:
            return
        self.load_variant(self.items[index])
        # self.last_variants.cache[self.parent_uuid] = variant_uuid

    # noinspection PyMethodMayBeStatic
    def on_activate_item(self, _):
        from ..launcherapp import LauncherApp

        LauncherApp.start_game()

    def on_setting(self, key, value):
        if key == "parent_uuid":
            self.parent_uuid = value
            if value:
                self.update_list(value)
            else:
                LauncherSettings.set("game_uuid", "")
                # self.set_items([gettext("Configuration")])
                self.set_items([])
                self.disable()
        elif key == "__variant_rating":
            variant_uuid = LauncherConfig.get("variant_uuid")
            for item in self.items:
                if item["uuid"] == variant_uuid:
                    item["personal_rating"] = int(value or 0)
                    self.update()
                    break

    def set_items(self, items):
        self.items = items
        self.update()
        # self.enable(len(items) > 0)

    def get_item_count(self):
        return len(self.items)

    def get_item_text(self, index):
        name = self.items[index]["name"]
        name = name.replace(", ", " \u00b7 ")
        return name

    def get_item_text_color(self, index):
        have = self.items[index]["have"]
        if not have:
            return self.missing_color

    NOT_AVAILABLE = 0
    MANUAL_AVAILABLE = 1
    AUTO_AVAILABLE = 2
    AVAILABLE = 3

    def get_item_icon(self, index):
        have = self.items[index]["have"]
        if have == self.NOT_AVAILABLE:
            # return self.missing_icon
            return self.blank_icon
        if have == self.MANUAL_AVAILABLE:
            return self.manual_download_icon
        if have == self.AUTO_AVAILABLE:
            return self.auto_download_icon
        return self.get_item_extra_icons(index)[0] or self.bullet_icon
        # name = self.items[index]["name"]
        # have = self.items[index]["have"]
        # if not have:
        #     return self.missing_icon
        # if have == 1:
        #     return self.manual_download_icon
        # if have == 2:
        #     return self.auto_download_icon
        #
        # if "IPF" in name:
        #     return self.ipf_icon
        # if "ADF" in name:
        #     return self.adf_icon
        # if "WHDLoad" in name:
        #     return self.hd_icon
        # if "CD32" in name:
        #     return self.cd_icon
        # if "CDTV" in name:
        #     return self.cd_icon
        # return self.icon

    def get_item_extra_icons(self, index):
        personal_rating = self.items[index]["personal_rating"]
        if personal_rating == 5:
            return [self.fav_icon]
        elif personal_rating == 4:
            return [self.up_icon]
        elif personal_rating == 3:
            return [self.up_icon]
        elif personal_rating == 2:
            return [self.down_icon]
        elif personal_rating == 1:
            return [self.down_icon]
        return [None]

    def update_list(self, game_uuid):
        print("VariantsBrowser.update_list, game_uuid=", game_uuid)
        database = Database.get_instance()
        items = database.find_game_variants_new(game_uuid, have=0)

        # items = database.search_configurations(self.search)
        # FIXME: Merge code with
        # FSGameSystemContext.py:get_ordered_game_variants
        sortable_items = []
        for i, variant in enumerate(items):
            name = variant["name"]
            # assert "\n" not in name
            # only show variant name (without game name)
            # name = name.split("\n", 1)[-1]

            game_database = fsgs.game_database(variant["database"])
            variant["like_rating"], variant["work_rating"] = (
                game_database.get_ratings_for_game(variant["uuid"])
            )
            variant["personal_rating"], ignored = (
                database.get_ratings_for_game(variant["uuid"])
            )

            if variant["published"] == 0:
                primary_sort = 1
                variant["name"] = "[UNPUBLISHED] " + variant["name"]
            else:
                primary_sort = 0
            sort_key = (
                primary_sort,
                1000000 - variant["like_rating"],
                1000000 - variant["work_rating"],
                name,
            )
            sortable_items.append((sort_key, i, variant))
        # print(sortable_items)
        self.items = [x[2] for x in sorted(sortable_items)]
        self.update()
        # self.set_items(self.items)
        # self.set_item_count(len(self.items))

        self.select_item(None, signal=False)

        select_index = None
        list_uuid = LauncherSettings.get("game_list_uuid")
        if list_uuid:
            list_variant_uuid = database.get_variant_for_list_and_game(
                list_uuid, game_uuid
            )
            print(
                "game list", list_uuid, "override variant", list_variant_uuid
            )
        else:
            list_variant_uuid = None
        if list_variant_uuid:
            # override variant selection from list if possible
            for i, variant in enumerate(self.items):
                print(variant["uuid"], variant["name"], list_variant_uuid)
                if variant["uuid"] == list_variant_uuid:
                    select_index = i
                    print("override select index", select_index)
                    break
        if select_index is None:
            # default index selection
            for i, variant in enumerate(self.items):
                if variant["personal_rating"] == 5:
                    if variant["have"] >= 3:
                        select_index = i
                        break
                    else:
                        # FIXME: Add warning here that the favorite variant
                        # could not be selected.
                        pass
            else:
                for i, variant in enumerate(self.items):
                    if variant["have"] >= 3:
                        select_index = i
                        break
                else:
                    for i, variant in enumerate(self.items):
                        if variant["have"] >= 1:
                            select_index = i
                            break
                    else:
                        if len(self.items) > 0:
                            select_index = 0

        # self.clear()
        # for i, variant in enumerate(self.items):
        #     self.add_item(variant["name"], icon=self.get_item_icon(i))

        self.enable(len(self.items) > 0)
        if select_index is not None:
            print("selecting variant index", select_index)
            self.select_item(select_index)
        else:
            ConfigGroup.new_config()

        # try:
        #     variant_uuid = self.last_variants.cache[game_uuid]
        # except KeyError:
        #     variant_uuid = database.get_last_game_variant(game_uuid)
        # if len(self.items) > 0:
        #     for i, variant in enumerate(self.items):
        #         if variant[3] == variant_uuid:
        #             self.select_item(i)
        #             break
        #     else:
        #         self.select_item(0)

    def load_variant(self, item):
        try:
            self._load_variant(item)
            # raise Exception()
        except Exception:
            traceback.print_exc()
            LauncherConfig.load_default_config()
            LauncherConfig.load({"__error": "Error Loading Configuration"})
            self.select_item(None)

    def _load_variant(self, item):
        print("_load_variant", item)
        variant_uuid = item["uuid"]
        database_name = item["database"]
        personal_rating = item["personal_rating"]
        have = item["have"]
        self._load_variant_2(
            variant_uuid, database_name, personal_rating, have
        )

    def _load_variant_2(
        self, variant_uuid, database_name, personal_rating, have
    ):
        if LauncherConfig.get("variant_uuid") == variant_uuid:
            print("Variant {} is already loaded".format(variant_uuid))
        game_database = fsgs.game_database(database_name)

        # game_database_client = GameDatabaseClient(game_database)
        # try:
        #     variant_id = game_database_client.get_game_id(variant_uuid)
        # except Exception:
        #     # game not found:
        #     print("could not find game", repr(variant_uuid))
        #     Config.load_default_config()
        #     return
        # values = game_database_client.get_final_game_values(variant_id)
        try:
            values = game_database.get_game_values_for_uuid(variant_uuid)
        except Exception:
            # game not found:
            traceback.print_exc()
            print("could not find game", repr(variant_uuid))
            LauncherConfig.load_default_config()
            return

        # values["variant_uuid"] = variant_uuid
        # values["variant_rating"] = str(item["personal_rating"])

        LauncherConfig.load_values(values, uuid=variant_uuid)

        # print("--->", LauncherConfig.get("variant_uuid"))

        # variant_rating = 0
        # if item["work_rating"] is not None:
        #     variant_rating = item["work_rating"] - 2
        # if item["like_rating"]:
        #     variant_rating = item["like_rating"]
        # Config.set("__variant_rating", str(variant_rating))

        # LauncherConfig.set("variant_uuid", variant_uuid)
        LauncherConfig.set("__changed", "0")
        LauncherConfig.set("__database", database_name)

        LauncherSettings.set("__variant_rating", str(personal_rating))

        if int(have) < self.AVAILABLE:
            print(" -- some files are missing --")
            LauncherConfig.set("x_missing_files", "1")

    def get_min_width(self):
        return 0



import fsui


class JoinDialog(fsui.LegacyDialog):
    def __init__(self):
        fsui.LegacyDialog.__init__(self, None, "Create/Join Game")
        self.layout = fsui.VerticalLayout()

        self.layout.add_spacer(20)

        hor_layout = fsui.HorizontalLayout()
        self.layout.add(hor_layout, fill=True)

        hor_layout.add_spacer(20)
        self.text = fsui.TextField(self)
        self.text.activated.connect(self.on_ok_button)
        hor_layout.add(self.text, expand=True)
        hor_layout.add_spacer(20)

        self.layout.add_spacer(20)

        hor_layout = fsui.HorizontalLayout()
        self.layout.add(hor_layout, fill=True)

        hor_layout.add_spacer(20, expand=True)
        self.cancel_button = fsui.Button(self, "    Cancel    ")
        self.cancel_button.activated.connect(self.on_cancel_button)
        hor_layout.add(self.cancel_button)
        hor_layout.add_spacer(10)
        self.ok_button = fsui.Button(self, "    Create/Join    ")
        self.ok_button.activated.connect(self.on_ok_button)
        hor_layout.add(self.ok_button)
        hor_layout.add_spacer(20)

        self.layout.add_spacer(20)
        self.set_size(self.layout.get_min_size())

        self.text.focus()

    def get_game_name(self):
        return "&" + self.text.get_text()

    def on_ok_button(self):
        self.end_modal(True)

    def on_cancel_button(self):
        self.end_modal(False)



import threading

from fsbc.signal import Signal as BaseSignal

main_thread_id = threading.current_thread().ident


class LauncherSignal(object):
    listeners = {}

    @classmethod
    def add_listener(cls, signal, listener):
        # cls.listeners.setdefault(signal, []).append(listener)
        BaseSignal(signal).connect(listener)

    @classmethod
    def remove_listener(cls, signal, listener):
        # cls.listeners[signal].remove(listener)
        BaseSignal(signal).disconnect(listener)

    @classmethod
    def broadcast(cls, signal, *args):
        if signal == "config":
            # temporary, while restructuring
            from fsgs.context import fsgs

            key, value = args
            fsgs.signal.notify("fsgs:config:" + key, value)

        # if threading.current_thread().ident == main_thread_id:
        #     cls.do_broadcast(signal, *args)
        # else:
        #     def function():
        #         cls.do_broadcast(signal, *args)
        #     fsui.call_after(function)
        BaseSignal(signal).notify(*args)



# Helper functions for fsbuild
# This file is automatically generated by fs-package

import datetime
import json
import locale
import os
import platform
import re
import shutil
import subprocess
import sys
import time
from os import path
from typing import Dict, List, Optional

macos = sys.platform == "darwin"
windows = sys.platform == "win32"
linux = sys.platform == "linux"

if windows:
    os_name = "Windows"
elif macos:
    os_name = "macOS"
elif linux:
    os_name = "Linux"
else:
    os_name = "Unknown"

machine = platform.machine()
if machine == "arm64" or machine == "aarch64":
    arch_name = "ARM64"
else:
    arch_name = "x86-64"


class PackageInformation:
    def __init__(self):
        self.values: Dict[str, str] = {}

    @property
    def macos_bundle_id(self) -> str:
        return self.values.get(
            "PACKAGE_BUNDLE_ID", self.values.get("PACKAGE_MACOS_BUNDLE_ID", "")
        )

    @property
    def display_name(self) -> str:
        """E.g. FS-UAE Launcher."""
        return self.values.get("PACKAGE_DISPLAY_NAME", self.pretty_name)

    @property
    def name(self) -> str:
        """E.g. fs-uae-launcher."""
        return self.values["PACKAGE_NAME"]

    @property
    def pretty_name(self) -> str:
        """E.g. FS-UAE-Launcher."""
        return self.values["PACKAGE_NAME_PRETTY"]

    @property
    def type(self):
        return self.values["PACKAGE_TYPE"]

    @property
    def version(self):
        return self.values["PACKAGE_VERSION"]

    @property
    def simple_version(self):
        return "{}.{}.{}".format(
            self.values["PACKAGE_VERSION_MAJOR"],
            self.values["PACKAGE_VERSION_MINOR"],
            self.values["PACKAGE_VERSION_REVISION"],
        )


_packageInformation: Optional[PackageInformation] = None


def getPackageInformation() -> PackageInformation:
    global _packageInformation
    if _packageInformation is None:
        _packageInformation = PackageInformation()
        with open("PACKAGE.FS", "r") as f:
            for line in f:
                try:
                    key, value = line.split("=", 1)
                    key = key.strip()
                    value = value.strip()
                    value = value.strip('"')
                    _packageInformation.values[key] = value
                except ValueError:
                    pass
    return _packageInformation


def getAppName() -> str:
    return getPackageInformation().display_name + ".app"


def getArchitecture() -> str:
    machine = platform.machine()
    if machine == "arm64":
        arch = "ARM64"
    else:
        arch = "x86-64"  # FIXME
    return arch


def getBundleName() -> str:
    packageInformation = getPackageInformation()
    if packageInformation.type == "fs-library-plugin":
        return getFrameworkName()
    else:
        return getAppName()


def getBundlePath(prefix: str = "build/_build/") -> str:
    # prettyName = getPackageInformation().pretty_name
    # bundleName = getBundleName()
    # arch = getArchitecture()
    # path = f"{prefix}{prettyName}/macOS/{arch}/{bundleName}"
    bundle_name = getBundleName()
    path = f"{prefix}/{bundle_name}"
    return path


def getDmgPath() -> str:
    packageInformation = getPackageInformation()
    prettyName = packageInformation.pretty_name
    version = packageInformation.version
    arch = getArchitecture()
    osDist = getOperatingSystemDist()
    path = f"build/_dist/{prettyName}_{version}_{osDist}_{arch}.dmg"
    return path


def getFrameworkName() -> str:
    return getPackageInformation().pretty_name + ".framework"


def getAppleTeamId() -> str:
    return os.environ.get("APPLE_TEAM_ID", "")


def getAppleId() -> str:
    return os.environ.get("APPLE_ID", "")


def getApplePassword() -> str:
    return os.environ.get("APPLE_PASSWORD", "")


def getAppleCodesignIdentity() -> str:
    return os.environ.get(
        "APPLE_CODESIGN_IDENTITY", "Developer ID Application"
    )


def getOperatingSystemDist() -> str:
    envValue = os.environ.get("SYSTEM_OS_DIST", "")
    if envValue:
        return envValue
    elif sys.platform == "linux":
        return "Linux"
    elif sys.platform == "darwin":
        return "macOS"
    elif sys.platform == "win32":
        return "Windows"
    return "Unknown"


def isMacOS() -> bool:
    return sys.platform == "darwin"


def notarizeApp(pathToNotarize: str, bundleId: str):
    assert path.exists(pathToNotarize)
    print(f"Notarizing {path.basename(pathToNotarize)}")
    result = shell(
        "xcrun notarytool submit "
        "--apple-id {appleId} "
        "--password {applePassword} "
        "--team-id {appleTeamId} "
        "--wait "
        "{pathToNotarize}".format(
            appleTeamId=getAppleTeamId(),
            appleId=getAppleId(),
            applePassword=getApplePassword(),
            pathToNotarize=pathToNotarize,
        ),
        censor=True,
    )
    print(result)
    if "status: Accepted" not in result:
        raise Exception("Notarization failed")


def run(args: List[str]) -> str:
    print(quoteArgs(args))
    return subprocess.check_output(args).decode("UTF-8")


def runCodeSign(args: List[str]) -> None:
    # Signing sometimes fails due to Apple errors (timeouts, etc). So we try
    # multiple times before giving up.
    for i in range(20):
        try:
            shell(quoteArgs(args))
        except Exception:
            time.sleep(1.0 * i)
            print("Attempt", i + 2)
        else:
            break
    else:
        raise Exception("Giving up signing")


def rmtree_if_exists(path: str) -> None:
    if os.path.exists(path):
        print("rmtree", path)
        shutil.rmtree(path)


def quoteArg(arg: str) -> str:
    if " " in arg:
        return f'"{arg}"'
    return arg


def quoteArgs(args: List[str]) -> str:
    return " ".join(f"{quoteArg(a)}" for a in args)


def shell(cmd: str, censor: bool = False) -> str:
    if censor:
        print("Running command (censored)")
    else:
        print(cmd)
    return subprocess.run(
        cmd, shell=True, check=True, stdout=subprocess.PIPE
    ).stdout.decode("UTF-8")


class Version:
    def __init__(self, version) -> None:
        self.version = version
        m = re.match("([0-9.]+[0-9])(.*)", version)
        assert m is not None
        parts = m.group(1).split(".")
        assert 2 <= len(parts) <= 4
        self.major = int(parts[0])
        self.minor = int(parts[1])
        self.build: Optional[int]
        if len(parts) > 3:
            self.build = int(parts[3])
        else:
            self.build = None
        if len(parts) > 2:
            self.revision = int(parts[2])
        else:
            self.revision = 0
        self.tag = m.group(2)
        self.commit = ""

    def setLast(self, n: int) -> None:
        if self.build is not None:
            self.build = n
        else:
            self.revision = n

    def __str__(self) -> str:
        numbers = [self.major, self.minor, self.revision]
        if self.build is not None:
            numbers.append(self.build)
        version = ".".join(str(x) for x in numbers)
        return version + self.tag


def numCommitsSince(base: str) -> int:
    to = "HEAD"
    result = int(
        subprocess.check_output(
            ["git", "rev-list", f"{base}..{to}", "--count"]
        ).decode()
    )
    return result


def findLastCommitForFile(path: str) -> str:
    commit = subprocess.check_output(
        ["git", "log", "-n", "1", "--pretty=format:%H", "--", path]
    ).decode()
    return commit


def findlastCommit() -> str:
    commit = subprocess.check_output(
        ["git", "log", "-n", "1", "--pretty=format:%H"]
    ).decode()
    return commit


def updateConfigureAc(version: Version, commit: str = "") -> None:
    print("Updating configure.ac")
    lines = []
    with open("configure.ac", "r", encoding="UTF-8") as f:
        for line in f:
            if line.startswith("m4_define([fsbuild_version"):
                if "_major" in line:
                    k = "FSBUILD_VERSION_MAJOR"
                    v = str(version.major)
                    # d = "Major version"
                elif "_minor" in line:
                    k = "FSBUILD_VERSION_MINOR"
                    v = str(version.minor)
                    # d = "Minor version"
                elif "_revision" in line:
                    k = "FSBUILD_VERSION_REVISION"
                    v = str(version.revision)
                    # d = "Revision"
                else:
                    k = "FSBUILD_VERSION"
                    v = str(version)
                    # d = "Full version"
                line = "m4_define([{}], [{}])\n".format(k.lower(), v)
            # if line.startswith("AC_DEFINE_UNQUOTED([FSBUILD_VERSION"):
            #     if "_MAJOR" in line:
            #         k = "FSBUILD_VERSION_MAJOR"
            #         v = version.major
            #         d = "Major version"
            #     elif "_MINOR" in line:
            #         k = "FSBUILD_VERSION_MINOR"
            #         v = version.minor
            #         d = "Minor version"
            #     elif "_REVISION" in line:
            #         k = "FSBUILD_VERSION_REVISION"
            #         v = version.revision
            #         d = "Revision"
            #     else:
            #         k = "FSBUILD_VERSION"
            #         v = str(version)
            #         d = "Full version"
            #     line = "AC_DEFINE_UNQUOTED([{}], [{}], [{}])\n".format(k, v, d)
            if line.startswith("m4_define([fsbuild_commit"):
                line = "m4_define([{}], [{}])\n".format(
                    "fsbuild_commit", commit
                )
            # if line.startswith("AC_DEFINE_UNQUOTED([FSBUILD_COMMIT"):
            #     k = "FSBUILD_COMMIT"
            #     v = commit
            #     d = "Package commit"
            #     line = "AC_DEFINE_UNQUOTED([{}], [{}], [{}])\n".format(k, v, d)
            lines.append(line)
    with open("configure.ac", "w", encoding="UTF-8") as f:
        for line in lines:
            f.write(line)


def updateDebianChangelog(version: Version) -> None:
    print("Updating debian/changelog")
    lines = []
    first_line = True
    first_line_changed = False
    deb_package = "unknown"
    deb_version = str(version)
    # deb_version = deb_version.replace("alpha", "~alpha")
    # deb_version = deb_version.replace("beta", "~beta")
    # deb_version = deb_version.replace("dev", "~dev")
    with open("debian/changelog", "r", encoding="UTF-8") as f:
        for line in f:
            if first_line:
                first_line = False
                deb_package = line.split(" ", 1)[0]
                lines.append(
                    "{} ({}-0) unstable; urgency=low\n".format(
                        deb_package, deb_version
                    )
                )
                if lines[-1] != line:
                    first_line_changed = True
            elif line.startswith(" -- ") and first_line_changed:
                # Only update date if version was changed
                author, date = line.split("  ")
                date = datetime.datetime.utcnow().strftime(
                    "%a, %d %b %Y %H:%M:%S +0000"
                )
                lines.append("{}  {}\n".format(author, date))
            else:
                lines.append(line)
    with open("debian/changelog", "w", encoding="UTF-8") as f:
        for line in lines:
            f.write(line)


def updateSpecFile(path: str, version: Version) -> None:
    print("Updating", path)
    lines = []
    rpm_version = str(version)
    # rpm_version = rpm_version.replace("alpha", "-0.1alpha")
    # rpm_version = rpm_version.replace("beta", "-0.1~beta")
    # rpm_version = rpm_version.replace("dev", "-0.1dev")
    # if not "-" in rpm_version:
    #     rpm_version += "-1"
    with open(path, "r", encoding="UTF-8") as f:
        for line in f:
            if line.startswith("%define fsbuild_version "):
                lines.append(
                    "%define fsbuild_version {}\n".format(rpm_version)
                )
            # elif line.startswith("%define unmangled_version "):
            #     lines.append("%define unmangled_version {0}\n".format(version))
            else:
                lines.append(line)
    with open(path, "w", newline="\n") as f:
        f.write("".join(lines))


def updatePackageFs(version: Version) -> None:
    print("Updating PACKAGE.FS")
    lines = []
    with open("PACKAGE.FS", "r", encoding="UTF-8") as f:
        for line in f:
            if line.startswith("PACKAGE_VERSION="):
                lines.append(f"PACKAGE_VERSION={str(version)}\n")
            elif line.startswith("PACKAGE_VERSION_MAJOR="):
                lines.append(f"PACKAGE_VERSION_MAJOR={str(version.major)}\n")
            elif line.startswith("PACKAGE_VERSION_MINOR="):
                lines.append(f"PACKAGE_VERSION_MINOR={str(version.minor)}\n")
            elif line.startswith("PACKAGE_VERSION_REVISION="):
                lines.append(
                    f"PACKAGE_VERSION_REVISION={str(version.revision)}\n"
                )
            elif line.startswith("PACKAGE_VERSION_TAG="):
                lines.append(f"PACKAGE_VERSION_TAG={str(version.tag)}\n")
            elif line.startswith("PACKAGE_COMMIT="):
                lines.append(f"PACKAGE_COMMIT={version.commit}\n")
            else:
                lines.append(line)
    with open("PACKAGE.FS", "w", newline="\n") as f:
        f.write("".join(lines))


def updateVersionFs(version: Version) -> None:
    print("Updating VERSION.FS")
    with open("VERSION.FS", "w") as f:
        f.write(str(version))
        f.write("\n")


def updateCommitFs(version: Version) -> None:
    print("Updating COMMIT.FS")
    with open("COMMIT.FS", "w") as f:
        if version.commit:
            f.write(version.commit)
            f.write("\n")


def updatePyProjectToml(version: Version) -> None:
    print("Updating pyproject.toml")
    with open("pyproject.toml", "r") as f:
        text = f.read()
    text = text.replace(
        'version = "0.0.0"',
        f'version = "{version.major}.{version.minor}.{version.revision}"',
    )
    with open("pyproject.toml", "w") as f:
        f.write(text)


def calculateVersion(
    auto_revision: bool = False,
    increment_revision: bool = False,
    include_commit: bool = False,
) -> Version:
    # with open("build/VERSION") as f:
    with open("BASEVERSION.FS") as f:
        version_str = f.read().strip()
    if version_str.startswith("BASEVERSION_FS="):
        version_str = version_str[len("BASEVERSION_FS=") :].strip()
    # with open("PACKAGE.FS") as f:
    #     for line in f:
    #         if line.startswith("PACKAGE_VERSION="):
    #             version_str = line[16:].strip()
    version = Version(version_str)
    if auto_revision:
        version_commit = findLastCommitForFile("BASEVERSION.FS")
        increment = numCommitsSince(version_commit)
        if increment_revision:
            increment += 1
        if version.build is not None:
            version.build += increment
        else:
            version.revision += increment
    if "--commit" in sys.argv:
        version.commit = findlastCommit()

    if True:
        branch = None
        githubRef = os.environ.get("GITHUB_REF")
        if githubRef is not None:
            if githubRef.startswith("refs/heads/"):
                branch = githubRef[len("refs/heads/") :]
            if githubRef.startswith("refs/pull/"):
                branch = "pull" + githubRef[len("refs/pull/") :].replace(
                    "/", ""
                )
        if not branch:
            branch = subprocess.check_output(
                ["git", "symbolic-ref", "--short", "HEAD"], encoding="UTF-8"
            ).strip()

        if branch == "stable":
            version.tag = ""
        elif branch:
            version.tag = f"-{branch}"
        else:
            raise Exception("Cannot calculate version tag from git ref")

    return version


def updateVersion(version: Version) -> None:
    if os.path.exists("VERSION.FS"):
        updateVersionFs(version)
    if os.path.exists("COMMIT.FS"):
        updateCommitFs(version)
    if os.path.exists("configure.ac"):
        updateConfigureAc(version)
    if os.path.exists("debian/changelog"):
        updateDebianChangelog(version)
    if os.path.exists("PACKAGE.FS"):
        updatePackageFs(version)
    if os.path.exists("pyproject.toml"):
        updatePyProjectToml(version)
    for filename in os.listdir("."):
        if filename.endswith(".spec"):
            updateSpecFile(filename, version)


def update_version_command() -> None:
    # For date/time formatting
    locale.setlocale(locale.LC_TIME, "C")

    # auto_revision = "--auto" in sys.argv
    auto_revision = True
    increment_revision = "--next" in sys.argv
    # include_commit = "--commit" in sys.argv
    include_commit = True
    # if "--auto-next" in sys.argv:
    #     auto_revision = True
    #     increment_revision = True
    version = calculateVersion(
        auto_revision=auto_revision,
        increment_revision=increment_revision,
        include_commit=include_commit,
    )
    for arg in sys.argv:
        if arg.startswith("--build="):
            version.build = int(arg[8:])
        elif arg.startswith("--revision="):
            version.revision = int(arg[11:])
        elif arg.startswith("--last="):
            version.setLast(int(arg[7:]))
        elif arg.startswith("--version="):
            version = Version(arg[10:])
    print(str(version))

    if "--print" in sys.argv:
        # Only print version
        pass
    else:
        updateVersion(version)


def buildDmg():
    if not isMacOS():
        print("Not building DMG on non-macOS platform")
        return
    dmgPath = getDmgPath()
    bundlePath = getBundlePath()
    packageInformation = getPackageInformation()
    print(f"Building {path.basename(dmgPath)}")
    if not path.exists(path.dirname(dmgPath)):
        os.makedirs(path.dirname(dmgPath))
    if os.path.exists(dmgPath):
        os.unlink(dmgPath)
    tool = "dmgbuild"
    if tool == "appdmg":
        bundlePath = getBundlePath(prefix="")
        settingsPath = "build/_build/appdmg.json"
        with open(settingsPath, "w", encoding="UTF-8") as f:
            json.dump(
                {
                    "title": packageInformation.display_name,
                    "contents": [
                        {
                            "x": 192,
                            "y": 344,
                            "type": "file",
                            "path": bundlePath,
                        },
                        {
                            "x": 448,
                            "y": 344,
                            "type": "link",
                            "path": "/Applications",
                        },
                    ],
                },
                f,
            )
        subprocess.check_call(
            [
                "appdmg",
                settingsPath,
                dmgPath,
            ]
        )
    elif tool == "dmgbuild":  # type: ignore
        bundlePath = getBundlePath()
        settingsPath = "build/dmgbuild/settings.py"
        # settingsPath = "build/_build/dmgbuild-settings.py"
        # with open(settingsPath, "w", encoding="UTF-8") as f:
        #     f.write("format = 'UDZO'\n")
        #     f.write("files = [\n")
        #     f.write(f"    '{bundlePath}',\n")
        #     f.write("]\n")
        #     f.write("symlinks = { 'Applications': '/Applications' }\n")
        #     f.write("badge_icon = 'icon/fs-uae-launcher.icns'\n")
        subprocess.check_call(
            [
                "dmgbuild",
                "-s",
                settingsPath,
                "-D",
                f'app="{bundlePath}"',
                "FS-UAE-Launcher",
                dmgPath,
            ]
        )
    else:
        raise Exception("Unknown dmg builder")


def signDmg():
    if not isMacOS():
        print("Not signing DMG on non-macOS platform")
        return
    args = [
        "codesign",
        "--force",
        "--sign",
        getAppleCodesignIdentity(),
        "--digest-algorithm=sha1,sha256",
    ]
    args.append(getDmgPath())
    runCodeSign(args)


def notarizeDmg():
    if not isMacOS():
        print("Not notarizing DMG on non-macOS platform")
        return
    packageInformation = getPackageInformation()
    bundleId = packageInformation.macos_bundle_id
    dmgPath = getDmgPath()
    assert path.exists(dmgPath)
    notarizeApp(dmgPath, bundleId)

    print('xcrun stapler staple "{}"'.format(dmgPath))
    assert os.system('xcrun stapler staple "{}"'.format(dmgPath)) == 0

    print("-" * 80)
    print(f"[BUILD] Notarized {dmgPath}")


def upload_branch_name():
    ref = os.getenv("GITHUB_REF", "")
    try:
        branch = ref.split("refs/heads/", 1)[1]
    except IndexError:
        return None
    if "/" in branch:
        return None
    return branch[0].upper() + branch[1:]



import fsui
from .application import Application


class Menu(fsui.Menu):
    def __init__(self, parent):
        if isinstance(parent, Application):
            parent.add_menu(self)
        else:
            print("FIXME: adding menu to window", parent)
        super().__init__()



from fsgs.drivers.messdriver import MessDriver


class MessNesDriver(MessDriver):
    CONTROLLER = {
        "type": "gamepad",
        "description": "Gamepad",
        "mapping_name": "nintendo",
    }

    PORTS = [
        {"description": "1st Controller", "types": [CONTROLLER]},
        {"description": "2nd Controller", "types": [CONTROLLER]},
    ]

    def mess_input_mapping(self, _):
        return {
            "START": "P#_START",
            "SELECT": "P#_SELECT",
            "UP": "P#_JOYSTICK_UP",
            "DOWN": "P#_JOYSTICK_DOWN",
            "LEFT": "P#_JOYSTICK_LEFT",
            "RIGHT": "P#_JOYSTICK_RIGHT",
            "A": "P#_BUTTON2",
            "B": "P#_BUTTON1",
        }

    def mess_configure(self):
        self.mess_configure_cartridge()

    def mess_offset_and_scale(self):
        # if self.get_romset() == "nespal":
        #     return 0.0, 0.0, 1.082, 1.250
        # return 0.0, 0.0, 1.072, 1.164
        return 0.0, 0.0, 1.072, 1.100

    def mess_romset(self):
        if self.get_game_refresh_rate() == 50:
            return "nespal", {}
        else:
            return "nes", {}



import fsui
from launcher.i18n import gettext
from launcher.ui.config.HardDriveGroup import HardDriveGroup
from launcher.ui.config.WHDLoadGroup import WHDLoadGroup
from launcher.ui.skin import Skin


class HardDrivesPanel(fsui.Panel):
    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        Skin.set_background_color(self)
        self.layout = fsui.VerticalLayout()

        heading_label = fsui.HeadingLabel(self, gettext("Hard Drives"))
        self.layout.add(heading_label, margin=10)
        self.layout.add_spacer(0)
        self.hard_drive_groups = []
        for i in range(4):
            self.hard_drive_groups.append(HardDriveGroup(self, i))
            self.layout.add(self.hard_drive_groups[i], fill=True, margin=10)

        self.layout.add_spacer(Skin.EXTRA_GROUP_MARGIN)
        heading_label = fsui.HeadingLabel(self, gettext("WHDLoad Arguments"))
        self.layout.add(heading_label, margin=10)
        self.layout.add_spacer(0)
        self.whdload_group = WHDLoadGroup(self)
        self.layout.add(self.whdload_group, fill=True)



import base64
import json
import platform
import time
from functools import wraps
from gzip import GzipFile
from io import StringIO
from urllib.error import HTTPError
from urllib.parse import urlencode
from urllib.request import Request
from uuid import uuid4

import requests

from fsbc.application import app
from fsbc.settings import Settings
from fsbc.task import Task
from fsgs.network import openretro_url_prefix


class NonRetryableHTTPError(HTTPError):
    pass


class BadRequestError(NonRetryableHTTPError):
    pass


class UnauthorizedError(NonRetryableHTTPError):
    pass


class ForbiddenError(NonRetryableHTTPError):
    pass


class NotFoundError(NonRetryableHTTPError):
    pass


def retry(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        for i in range(10):
            if i > 0:
                print("retrying...")
            try:
                return f(*args, **kwargs)
            except NonRetryableHTTPError as e:
                raise e
            except Exception as e:
                print(repr(e))
                time.sleep(i * 0.5)
        return f(*args, **kwargs)

    return wrapper


class OGDClient(object):
    HTTPError = HTTPError
    BadRequestError = BadRequestError
    UnauthorizedError = UnauthorizedError
    ForbiddenError = ForbiddenError
    NotFoundError = NotFoundError
    NonRetryableHTTPError = NonRetryableHTTPError

    def __init__(self):
        self._json = None
        self.data = b""

    @staticmethod
    def is_logged_in():
        # non-empty ogd_auth means we are logged in (probably, the
        # authorization can in theory have been invalidated on the server
        return bool(Settings.instance()["database_auth"])

    def login_task(self, username, password):
        return LoginTask(self, username, password)

    def logout_task(self, auth_token):
        return LogoutTask(self, auth_token)

    @retry
    def authorize(self, username, password, device_id, device_name):
        result = self.post(
            "/api/auth",
            {
                "username": username,
                "password": password,
                "device_id": device_id,
                "device_name": device_name,
            },
            auth=False,
        )
        return result

    @retry
    def deauth(self, auth_token):
        result = self.post(
            "/api/deauth", {"auth_token": auth_token}, auth=False
        )
        return result

    @staticmethod
    def url_prefix():
        return openretro_url_prefix()

    @staticmethod
    def auth():
        auth_token = Settings.instance()["database_auth"]
        return ("auth_token", auth_token)

    def post(self, path, params=None, data=None, auth=True):
        headers = {}
        url = "{0}{1}".format(openretro_url_prefix(), path)
        if not data and params:
            data = urlencode(params)
            headers["Content-Type"] = "application/x-www-form-urlencoded"
        print(url, headers)
        r = requests.post(
            url, data, headers=headers, auth=(self.auth() if auth else None)
        )
        r.raise_for_status()
        return r.json()

    def build_url(self, path, **kwargs):
        url = "{0}{1}".format(self.url_prefix(), path)
        if kwargs:
            url += "?" + urlencode(kwargs)
        return url

    def rate_variant(self, variant_uuid, like=None, work=None):
        params = {"game": variant_uuid}
        if like is not None:
            params["like"] = like
        if work is not None:
            params["work"] = work
        url = self.build_url("/api/1/rate_game", **params)
        r = requests.get(url, auth=self.auth())
        r.raise_for_status()
        return r.json()


def get_device_name():
    try:
        return platform.node() or "Unknown Computer"
    except Exception:
        return "Unknown Computer"


class LoginTask(Task):
    def __init__(self, client, username, password):
        Task.__init__(self, "Login Task")
        self.client = client
        self.username = username
        self.password = password

    def run(self):
        self.progressed("Logging into openretro.org...")
        if not Settings.instance()["device_id"]:
            Settings.instance()["device_id"] = str(uuid4())
        try:
            result = self.client.authorize(
                self.username,
                self.password,
                Settings.instance()["device_id"],
                get_device_name(),
            )
        except UnauthorizedError:
            raise Task.Failure("Wrong e-mail address or password")

        Settings.instance()["database_username"] = result["username"]
        Settings.instance()["database_email"] = result["email"]
        Settings.instance()["database_auth"] = result["auth_token"]
        Settings.instance()["database_password"] = ""


class LogoutTask(Task):
    def __init__(self, client, auth_token):
        Task.__init__(self, "Logout Task")
        self.client = client
        self.auth_token = auth_token

    def run(self):
        self.progressed("Logging out from openretro.org...")
        if not Settings.instance()["device_id"]:
            Settings.instance()["device_id"] = str(uuid4())
        self.client.deauth(self.auth_token)

        Settings.instance()["database_username"] = ""
        # Settings.instance()["database_email"] = ""
        Settings.instance()["database_auth"] = ""
        Settings.instance()["database_password"] = ""



import logging

from fsbc.resources import Resources as BaseResources

logger = logging.getLogger("arcade")


def _(msg):
    return str(msg)


def ngettext(n, msg1, msg2):
    return str(msg1) if n == 1 else str(msg2)


class Resources(BaseResources):
    def __init__(self, package):
        BaseResources.__init__(self, package, "res")
        self.req = package

    def resource_pil_image(self, resource):
        # resource_name = encode_path(u'res/' + resource_name)
        # return resource_pil_image(self.req, resource_name)
        stream = self.stream(resource)
        # noinspection PyUnresolvedReferences
        from PIL import Image

        return Image.open(stream)

    def resource_qt_image(self, resource):
        from fsui.qt import QImage

        stream = self.stream(resource)
        im = QImage()
        im.loadFromData(stream.read())
        return im

    def resource_stream(self, resource):
        return self.stream(resource)

    def resource_filename(self, resource):
        return self.path(resource)


gettext = _

resources = Resources("arcade")



import fsui
from ..common.group import Group


class IconHeader(Group):
    ICON_LEFT = 0
    ICON_RIGHT = 1

    def __init__(
        self, parent, icon, title, subtitle="", icon_position=ICON_LEFT
    ):
        Group.__init__(self, parent)
        self.layout = fsui.HorizontalLayout()

        image = icon.image(48)
        self.image_view = fsui.ImageView(self, image)
        if icon_position == self.ICON_LEFT:
            self.layout.add(self.image_view)
            self.layout.add_spacer(20)
        vert_layout = fsui.VerticalLayout()
        self.layout.add(vert_layout, expand=True, fill=False, valign=0.5)
        self.title_label = fsui.HeadingLabel(self, title)
        vert_layout.add(self.title_label)
        vert_layout.add_spacer(2)
        self.subtitle_label = fsui.Label(self, subtitle)
        vert_layout.add(self.subtitle_label, fill=True)
        if icon_position == self.ICON_RIGHT:
            self.layout.add_spacer(20)
            self.layout.add(self.image_view)


class NewIconHeader(Group):
    ICON_LEFT = 0
    ICON_RIGHT = 1

    def __init__(
        self, parent, icon, title, subtitle="", icon_position=ICON_RIGHT
    ):
        Group.__init__(self, parent)
        self.layout = fsui.HorizontalLayout()

        image = icon.image(48)
        self.image_view = fsui.ImageView(self, image)
        if icon_position == self.ICON_LEFT:
            self.layout.add(self.image_view)
            self.layout.add_spacer(20)

        # vert_layout = fsui.VerticalLayout()
        # self.layout.add(
        #     vert_layout, expand=True, fill=False, valign=0.5)
        self.title_label = fsui.HeadingLabel(self, title)
        font = self.title_label.get_font()
        font.increase_size(3)
        self.title_label.set_font(font)

        self.layout.add(self.title_label, expand=True, fill=False, valign=0.5)
        if icon_position == self.ICON_RIGHT:
            self.layout.add_spacer(20)
            self.layout.add(self.image_view)



import fsui
from launcher.i18n import gettext
from launcher.option import Option
from launcher.settings.settings_page import SettingsPage


class ExperimentalFeaturesPage(SettingsPage):
    def __init__(self, parent):
        super().__init__(parent)
        icon = fsui.Icon("settings", "pkg:workspace")
        title = gettext("Experimental Features")
        subtitle = ""
        self.add_header(icon, title, subtitle)

        self.add_option(Option.NETPLAY_FEATURE)
        self.add_option(Option.LAUNCHER_CONFIG_FEATURE)



import fsui
from fsbc.system import linux
from fsgs.option import Option
from launcher.res import gettext
from launcher.settings.settings_page import SettingsPage


class FSUAESettingsPage(SettingsPage):
    def __init__(self, parent):
        super().__init__(parent)
        icon = fsui.Icon("fs-uae", "pkg:launcher")
        self.add_header(icon, "FS-UAE")

        label = fsui.MultiLineLabel(
            self,
            gettext(
                "The following options may affect performance, "
                "so only enable them when needed for testing or "
                "debugging purposes."
            ),
            640,
        )
        self.layout.add(label, fill=True, margin_top=20)

        if linux:
            self.add_option(Option.GAME_MODE)
            self.add_option(Option.GOVERNOR_WARNING)

        self.add_option(Option.LOG_AUTOSCALE)
        self.add_option(Option.LOG_INPUT)



from __future__ import absolute_import
from __future__ import print_function

import sys

class Progress:
  def __init__(self):
    self.num = 0  
    self.msg = None
  def begin(self, msg):
    self.num = 0
    self.msg = msg
  def end(self):
    pass
  def add(self):
    self.num += 1